"""Support code for production hardware tests

We execute the production (hardware tests) for new hardware. Usually this
will be one of the following hardware units:

- stationary transceiver unit
- sensory tool holder (STH)/sensory tool assembly
- sensory milling head (SMH)
"""



"""Support for ADC CAN commands and configuration

See:
https://mytoolit.github.io/Documentation/#command:Get-Set-ADC-Configuration

for more information
"""

# -- Imports ------------------------------------------------------------------

from collections.abc import Iterator, Mapping
from math import log2

from icotronic.utility.types import check_list

# -- Class --------------------------------------------------------------------


class ADCConfiguration(Mapping):
    """Support for reading and writing analog digital converter configuration

    Args:
        *data:
            A list containing the (first five) bytes of the ADC
            configuration

        set:
            Specifies if we want to set or retrieve (get) the ADC
            configuration

        prescaler:
            The ADC prescaler value (1 – 127)

        acquisition_time:
            The acquisition time in number of cycles
            (1, 2, 3, 4, 8, 16, 32, … , 256)

        oversampling_rate:
            The ADC oversampling rate (1, 2, 4, 8, … , 4096)

        reference_voltage:
            The ADC reference voltage in Volt
            (1.25, 1.65, 1.8, 2.1, 2.2, 2.5, 2.7, 3.3, 5, 6.6)

    Examples:

        Create simple ADC configuration from scratch

        >>> ADCConfiguration(prescaler=2,
        ...     acquisition_time=4,
        ...     oversampling_rate=64) # doctest:+NORMALIZE_WHITESPACE
        Get, Prescaler: 2, Acquisition Time: 4, Oversampling Rate: 64,
        Reference Voltage: 3.3 V

    """

    REFERENCE_VOLTAGES = [1.25, 1.65, 1.8, 2.1, 2.2, 2.5, 2.7, 3.3, 5, 6.6]

    # pylint: disable=too-many-branches

    def __init__(  # pylint: disable=too-many-arguments
        self,
        *data: bytearray | list[int],
        # pylint: disable=redefined-builtin
        set: bool | None = None,
        # pylint: enable=redefined-builtin
        prescaler: int | None = None,
        acquisition_time: int | None = None,
        oversampling_rate: int | None = None,
        reference_voltage: float | None = None,
    ):
        if data:
            data_bytes = list(data[0])
            check_list(data_bytes, 5)
            self.data = data_bytes[0:5] + [0] * 3
        else:
            self.data = [0] * 8

        # ==================
        # = Get/Set Config =
        # ==================

        if set is not None:
            get_set_byte = self.data[0]
            # Set get/set to 0
            get_set_byte &= 0b01111111
            # Set value
            get_set_byte |= int(set) << 7
            self.data[0] = get_set_byte

        # =============
        # = Prescaler =
        # =============

        if prescaler is not None:
            self.prescaler = prescaler
        elif self.data[1] == 0:
            # Make sure default prescaler value makes sense
            self.prescaler = 8

        # ====================
        # = Acquisition Time =
        # ====================

        if acquisition_time is not None:
            self.acquisition_time = acquisition_time

        # =====================
        # = Oversampling Rate =
        # =====================

        if oversampling_rate is not None:
            self.oversampling_rate = oversampling_rate

        # =====================
        # = Reference Voltage =
        # =====================

        if reference_voltage is not None:
            self.reference_voltage = reference_voltage
        elif self.data[4] == 0:
            # Make sure default reference voltage value makes sense
            self.reference_voltage = 3.3

        self.attributes = {
            "reference_voltage": self.reference_voltage,
            "prescaler": self.prescaler,
            "acquisition_time": self.acquisition_time,
            "oversampling_rate": self.oversampling_rate,
        }

    # pylint: enable=too-many-branches

    def __getitem__(self, item: str) -> float:
        """Return values of the mapping provided by this class

        Note:
            This method allow access to the object via the splat operators
            (*, **)

        Args:

            item:
                The attribute for which we want to retrieve the value

        Returns:

            The value of the attribute

        Examples:

            Convert ADC configurations into dictionary

            >>> dict(**ADCConfiguration()) # doctest:+NORMALIZE_WHITESPACE
            {'reference_voltage': 3.3,
             'prescaler': 8,
             'acquisition_time': 1,
             'oversampling_rate': 1}

            >>> dict(**ADCConfiguration(oversampling_rate=64)
            ...     ) # doctest:+NORMALIZE_WHITESPACE
            {'reference_voltage': 3.3,
             'prescaler': 8,
             'acquisition_time': 1,
             'oversampling_rate': 64}

        """

        return self.attributes[item]

    def __iter__(self) -> Iterator:
        """Return an iterator over the mapping provided by this class

        Note:

            This method allow access to the object via the splat operators
            (*, **)

        Returns:

            The names of the “important” properties of the ADC configuration:

            - reference voltage
            - prescaler
            - acquisition time
            - oversampling rate

        Examples:

            Print ADC attribute keys

            >>> for attribute in ADCConfiguration():
            ...     print(attribute)
            reference_voltage
            prescaler
            acquisition_time
            oversampling_rate

        """

        return iter(self.attributes)

    def __len__(self) -> int:
        """Return the length of the mapping provided by this class

        Note:

            This method allow access to the object via the splat operators
            (*, **)

        Returns:

            The amount of the “important” properties of the ADC configuration:

            - reference voltage
            - prescaler
            - acquisition time
            - oversampling rate

        Examples:

            Get the (constant) length of the ADC configuration

            >>> len(ADCConfiguration())
            4

            >>> len(ADCConfiguration(reference_voltage=3.3))
            4

        """

        return len(self.attributes)

    def __repr__(self) -> str:
        """Retrieve the textual representation of the ADC configuration

        Returns:

            A string that describes the ADC configuration

        Examples:

            Retrieve the string representation of ADC configurations

            >>> ADCConfiguration(prescaler=1, reference_voltage=3.3
            ... ) # doctest:+NORMALIZE_WHITESPACE
            Get, Prescaler: 1, Acquisition Time: 1, Oversampling Rate: 1,
            Reference Voltage: 3.3 V

            >>> ADCConfiguration(
            ...     set=True,
            ...     prescaler=64,
            ...     acquisition_time=128,
            ...     oversampling_rate=1024,
            ...     reference_voltage=1.8) # doctest:+NORMALIZE_WHITESPACE
            Set, Prescaler: 64, Acquisition Time: 128, Oversampling Rate: 1024,
            Reference Voltage: 1.8 V

            >>> ADCConfiguration(
            ...     [0, 2, 4, 6, 25]) # doctest:+NORMALIZE_WHITESPACE
            Get, Prescaler: 2, Acquisition Time: 8, Oversampling Rate: 64,
            Reference Voltage: 1.25 V

        """

        set_values = bool(self.data[0] >> 7)

        parts = [
            "Set" if set_values else "Get",
            f"Prescaler: {self.prescaler}",
            f"Acquisition Time: {self.acquisition_time}",
            f"Oversampling Rate: {self.oversampling_rate}",
            f"Reference Voltage: {self.reference_voltage} V",
        ]

        return ", ".join(parts)

    @property
    def reference_voltage(self) -> float:
        """Get the reference voltage

        Returns:

            The reference voltage in Volt

        Examples:

            Set and get different reference voltage values

            >>> config = ADCConfiguration(reference_voltage=3.3)
            >>> config.reference_voltage
            3.3
            >>> config.reference_voltage = 6.6
            >>> config.reference_voltage
            6.6

            >>> config = ADCConfiguration(reference_voltage=6.6)
            >>> config.reference_voltage
            6.6
            >>> config.reference_voltage = 1.8
            >>> config.reference_voltage
            1.8

            >>> config = ADCConfiguration(reference_voltage=1.8)
            >>> config.reference_voltage
            1.8

            Trying to set a unsupported reference voltage will fail

            >>> config.reference_voltage = 0 # doctest:+ELLIPSIS
            Traceback (most recent call last):
               ...
            ValueError: Reference voltage of “0” V out of range, please use ...

        """

        return self.data[4] / 20

    @reference_voltage.setter
    def reference_voltage(self, reference_voltage: float) -> None:
        """Change the reference voltage

        Args:

            reference_voltage:
                The new value for the reference voltage in V

        """

        cls = type(self)
        if reference_voltage not in cls.REFERENCE_VOLTAGES:
            raise ValueError(
                f"Reference voltage of “{reference_voltage}” V out of range"
                ", please use one of the following values: "
                + ", ".join(map(str, cls.REFERENCE_VOLTAGES))
            )

        self.data[4] = int(reference_voltage * 20)

    @property
    def prescaler(self) -> int:
        """Get the prescaler value

        Returns:

            The prescaler value

        Examples:

            Get initialised prescaler

            >>> config = ADCConfiguration(prescaler=127)
            >>> config.prescaler
            127

            Get prescaler set via property

            >>> config.prescaler = 20
            >>> config.prescaler
            20

            Trying to set an unsupported prescaler will fail

            >>> config.prescaler = 128 # doctest:+ELLIPSIS
            Traceback (most recent call last):
               ...
            ValueError: Prescaler value of “128” out of range, please use ...

        """

        return self.data[1]

    @prescaler.setter
    def prescaler(self, prescaler: int) -> None:
        """Change the prescaler value

        Args:

            prescaler:
                The new value for the prescaler

        """

        if not 1 <= prescaler <= 127:
            raise ValueError(
                f"Prescaler value of “{prescaler}” out of range"
                ", please use a value between 1 and 127"
            )
        self.data[1] = prescaler

    @property
    def acquisition_time(self) -> int:
        """Get the acquisition time

        Returns:

            The acquisition time

        Examples:

            Get initialised acquisition time

            >>> config = ADCConfiguration(acquisition_time=2)
            >>> config.acquisition_time
            2

            Get acquisition time set via property

            >>> config.acquisition_time = 128
            >>> config.acquisition_time
            128

            Trying to set an unsupported acquisition time will fail

            >>> config.acquisition_time = 5 # doctest:+ELLIPSIS
            Traceback (most recent call last):
               ...
            ValueError: Acquisition time of “5” out of range, please use ...

        """

        acquisition_time_byte = self.data[2]

        return (
            acquisition_time_byte + 1
            if acquisition_time_byte <= 3
            else 2 ** (acquisition_time_byte - 1)
        )

    @acquisition_time.setter
    def acquisition_time(self, acquisition_time: int) -> None:
        """Change the acquisition time value

        Args:

            acquisition_time:
                The new value for the acquisition time

        """

        possible_acquisition_times = list(range(1, 5)) + [
            2**value for value in range(3, 9)
        ]
        if acquisition_time not in possible_acquisition_times:
            raise ValueError(
                f"Acquisition time of “{acquisition_time}” out of range"
                ", please use one of the following values: "
                + ", ".join(map(str, possible_acquisition_times))
            )

        acquisition_time_byte = (
            acquisition_time - 1
            if acquisition_time <= 3
            else int(log2(acquisition_time)) + 1
        )

        self.data[2] = acquisition_time_byte

    @property
    def oversampling_rate(self) -> int:
        """Get the oversampling rate

        Returns:

            The oversampling rate

        Examples:

            Get initialised oversampling rate

            >>> config = ADCConfiguration(oversampling_rate=128)
            >>> config.oversampling_rate
            128

            Get oversampling rate set via property

            >>> config.oversampling_rate = 512
            >>> config.oversampling_rate
            512

            Trying to set an unsupported oversampling rate will fail

            >>> config.oversampling_rate = 3 # doctest:+ELLIPSIS
            Traceback (most recent call last):
               ...
            ValueError: Oversampling rate of “3” out of range, please use ...

        """

        oversampling_rate_byte = self.data[3]

        return 2**oversampling_rate_byte

    @oversampling_rate.setter
    def oversampling_rate(self, oversampling_rate: int) -> None:
        """Change the oversampling rate

        Args:

            oversampling_rate:
                The new value for the oversampling rate

        """

        possible_oversampling_rates = [2**value for value in range(13)]
        if oversampling_rate not in possible_oversampling_rates:
            raise ValueError(
                f"Oversampling rate of “{oversampling_rate}” out of "
                "range, please use one of the following values: "
                + ", ".join(map(str, possible_oversampling_rates))
            )

        self.data[3] = int(log2(oversampling_rate))

    def sample_rate(self) -> float:
        """Calculate the sampling rate for the current ADC configuration

        Returns:

            The calculated sample rate

        Examples:

            Get sampling rates based on ADC attribute values

            >>> round(ADCConfiguration(prescaler=2, acquisition_time=8,
            ...                        oversampling_rate=64).sample_rate())
            9524

            >>> round(ADCConfiguration(prescaler=8, acquisition_time=8,
            ...                        oversampling_rate=64).sample_rate())
            3175

            >>> round(ADCConfiguration(reference_voltage=5.0,
            ...                        prescaler=16,
            ...                        acquisition_time=8,
            ...                        oversampling_rate=128).sample_rate())
            840

        """

        clock_frequency = 38_400_000

        return clock_frequency / (
            (self.prescaler + 1)
            * (self.acquisition_time + 13)
            * self.oversampling_rate
        )


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



"""Node specific EEPROM support"""

# -- Imports ------------------------------------------------------------------

from datetime import date

from semantic_version import Version

from icotronic.can.node.eeprom.basic import EEPROM
from icotronic.can.node.eeprom.status import EEPROMStatus

# -- Classes ------------------------------------------------------------------

# pylint: disable=too-many-public-methods


class NodeEEPROM(EEPROM):
    """Read and write node specific EEPROM data (STU/sensor nodes)"""

    # ========================
    # = System Configuration =
    # ========================

    async def read_status(self) -> EEPROMStatus:
        """Retrieve EEPROM status byte

        Returns:

            An EEPROM status object for the current status byte value

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Read the status byte of STU 1

            >>> async def read_status_byte():
            ...     async with Connection() as stu:
            ...         return await stu.eeprom.read_status()
            >>> isinstance(run(read_status_byte()), EEPROMStatus)
            True

        """

        return EEPROMStatus(
            (await self.read(address=0, offset=0, length=1)).pop()
        )

    async def write_status(self, value: int | EEPROMStatus) -> None:
        """Change the value of the EEPROM status byte

        Args:

            value:
                The new value for the status byte


        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Write and read the status byte of STU 1

            >>> async def write_read_status_byte():
            ...     async with Connection() as stu:
            ...         await stu.eeprom.write_status(
            ...             EEPROMStatus('Initialized'))
            ...         return await stu.eeprom.read_status()
            >>> status = run(write_read_status_byte())
            >>> status.is_initialized()
            True

        """

        await self.write_int(
            address=0, offset=0, length=1, value=EEPROMStatus(value).value
        )

    async def read_name(self) -> str:
        """Retrieve the name of the node from the EEPROM

        Returns:

            The name of the node

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Read the name of STU 1

            >>> async def read_name():
            ...     async with Connection() as stu:
            ...         return await stu.eeprom.read_name()
            >>> isinstance(run(read_name()), str)
            True

        """

        return await self.read_text(address=0, offset=1, length=8)

    async def write_name(self, name: str) -> None:
        """Write the name of the node into the EEPROM

        Args:

            name:
                The new (Bluetooth advertisement) name of the node

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Write and read the name of STU 1

            >>> async def write_read_name(name):
            ...     async with Connection() as stu:
            ...         await stu.eeprom.write_name(name)
            ...         return await stu.eeprom.read_name()
            >>> run(write_read_name('Valerie'))
            'Valerie'

        """

        await self.write_text(address=0, offset=1, text=name, length=8)

    # ================
    # = Product Data =
    # ================

    async def read_gtin(self) -> int:
        """Read the global trade identifier number (GTIN) from the EEPROM

        Returns:

            The GTIN of the specified node

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Read the GTIN of STU 1

            >>> async def read_gtin():
            ...     async with Connection() as stu:
            ...         return await stu.eeprom.read_gtin()
            >>> gtin = run(read_gtin())
            >>> isinstance(gtin, int)
            True

        """

        return await self.read_int(address=4, offset=0, length=8)

    async def write_gtin(self, gtin: int) -> None:
        """Write the global trade identifier number (GTIN) to the EEPROM

        Args:

            gtin:
                The new GTIN of the specified receiver

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Write and read the GTIN of STU 1

            >>> async def write_read_gtin(gtin):
            ...     async with Connection() as stu:
            ...         await stu.eeprom.write_gtin(gtin=gtin)
            ...         return await stu.eeprom.read_gtin()
            >>> run(write_read_gtin(0))
            0

        """

        await self.write_int(address=4, offset=0, length=8, value=gtin)

    async def read_hardware_version(self) -> Version:
        """Read the current hardware version from the EEPROM

        Returns:

            The hardware version of the node

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Read the hardware version of STU 1

            >>> async def read_hardware_version():
            ...     async with Connection() as stu:
            ...         return (await stu.eeprom.read_hardware_version())
            >>> hardware_version = run(read_hardware_version())
            >>> hardware_version.major >= 1
            True

        """

        major, minor, patch = await self.read(address=4, offset=13, length=3)
        return Version(major=major, minor=minor, patch=patch)

    async def write_hardware_version(self, version: str | Version):
        """Write hardware version to the EEPROM

        Args:

            version:
                The new hardware version of the node

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Write and read the hardware version of STU 1

            >>> async def write_read_hardware_version(version):
            ...     async with Connection() as stu:
            ...         await stu.eeprom.write_hardware_version(version)
            ...         return (await stu.eeprom.read_hardware_version())
            >>> hardware_version = run(write_read_hardware_version('1.3.2'))
            >>> hardware_version.patch == 2
            True

        """

        if isinstance(version, str):
            version = Version(version)

        await self.write(
            address=4,
            offset=13,
            length=3,
            data=[version.major, version.minor, version.patch],
        )

    async def read_firmware_version(self) -> Version:
        """Retrieve the current firmware version from the EEPROM

        Returns:

            The firmware version of the node

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Read the firmware version of STU 1

            >>> async def read_firmware_version():
            ...     async with Connection() as stu:
            ...         return await stu.eeprom.read_firmware_version()
            >>> firmware_version = run(read_firmware_version())
            >>> firmware_version.major >= 2
            True

        """

        major, minor, patch = await self.read(address=4, offset=21, length=3)
        return Version(major=major, minor=minor, patch=patch)

    async def write_firmware_version(self, version: str | Version) -> None:
        """Write firmware version to the EEPROM

        Args:

            version:
                The new firmware version

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Write and read the firmware version of STU 1

            >>> async def write_read_firmware_version(version):
            ...     async with Connection() as stu:
            ...         await stu.eeprom.write_firmware_version(version)
            ...         return (await stu.eeprom.read_firmware_version())
            >>> version = '2.1.10'
            >>> firmware_version = run(write_read_firmware_version(version))
            >>> firmware_version == Version(version)
            True

        """

        if isinstance(version, str):
            version = Version(version)

        await self.write(
            address=4,
            offset=21,
            length=3,
            data=[version.major, version.minor, version.patch],
        )

    async def read_release_name(self) -> str:
        """Retrieve the current release name from the EEPROM

        Returns:

            The firmware release name of the specified node

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Read the release name of STU 1

            >>> async def read_release_name():
            ...     async with Connection() as stu:
            ...         return await stu.eeprom.read_release_name()
            >>> run(read_release_name())
            'Valerie'

        """

        return await self.read_text(address=4, offset=24, length=8)

    async def write_release_name(self, name: str):
        """Write the release name to the EEPROM

        Args:

            name:
                The new name of the release

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Write and read the release name of STU 1

            >>> async def write_read_release_name(name):
            ...     async with Connection() as stu:
            ...         await stu.eeprom.write_release_name(name)
            ...         return await stu.eeprom.read_release_name()
            >>> run(write_read_release_name('Valerie'))
            'Valerie'

        """

        await self.write_text(address=4, offset=24, length=8, text=name)

    async def read_serial_number(self) -> str:
        """Retrieve the serial number from the EEPROM

        Returns:

            The serial number of the node

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Read the serial number of STU 1

            >>> async def read_serial_number():
            ...     async with Connection() as stu:
            ...         return await stu.eeprom.read_serial_number()
            >>> serial_number = run(read_serial_number())
            >>> isinstance(serial_number, str)
            True

        """

        return await self.read_text(address=4, offset=32, length=32)

    async def write_serial_number(self, serial_number: str):
        """Write the serial number to the EEPROM

        Args:

            serial_number:
                The serial number of the specified receiver

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Write and read the serial number of STU 1

            >>> async def write_read_serial_number(serial):
            ...     async with Connection() as stu:
            ...         await stu.eeprom.write_serial_number(serial)
            ...         return await stu.eeprom.read_serial_number()
            >>> run(write_read_serial_number('0'))
            '0'

        """

        await self.write_text(
            address=4, offset=32, length=32, text=serial_number
        )

    async def read_product_name(self) -> str:
        """Retrieve the product name from the EEPROM

        Returns:

            The product name of the node

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Read the product name of STU 1

            >>> async def read_product_name():
            ...     async with Connection() as stu:
            ...         return await stu.eeprom.read_product_name()
            >>> product_name = run(read_product_name())
            >>> isinstance(product_name, str)
            True

        """

        return await self.read_text(address=4, offset=64, length=128)

    async def write_product_name(self, name: str):
        """Write the product name to the EEPROM

        Args:

            name:
                The new product name of the specified receiver

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Write and read the product name of STU 1

            >>> async def write_read_product_name(name):
            ...     async with Connection() as stu:
            ...         await stu.eeprom.write_product_name(name)
            ...         return await stu.eeprom.read_product_name()
            >>> run(write_read_product_name('0'))
            '0'

        """

        await self.write_text(address=4, offset=64, length=128, text=name)

    async def read_oem_data(self) -> list[int]:
        """Retrieve the OEM data from the EEPROM

        Returns:

            The OEM data of the specified node

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Read the OEM data of STU 1

            >>> async def read_oem_data():
            ...     async with Connection() as stu:
            ...         return await stu.eeprom.read_oem_data()
            >>> oem_data = run(read_oem_data())
            >>> len(oem_data) == 64
            True

        """

        return await self.read(address=4, offset=192, length=64)

    async def write_oem_data(self, data: list[int]):
        """Write OEM data to the EEPROM

        Args:

            data:
                The OEM data that should be stored in the EEPROM

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Write and read the OEM data of STU 1

            >>> async def write_read_oem_data(data):
            ...     async with Connection() as stu:
            ...         await stu.eeprom.write_oem_data(data)
            ...         return await stu.eeprom.read_oem_data()
            >>> data = [0] * 64
            >>> run(write_read_oem_data(data)) == data
            True

        """

        await self.write(address=4, offset=192, length=64, data=data)

    # ==============
    # = Statistics =
    # ==============

    async def read_power_on_cycles(self) -> int:
        """Retrieve the number of power on cycles from the EEPROM

        Returns:

            The number of power on cycles of the node

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Read the number of power on cycles of STU 1

            >>> async def read_power_on_cycles():
            ...     async with Connection() as stu:
            ...         return await stu.eeprom.read_power_on_cycles()
            >>> power_on_cycles = run(read_power_on_cycles())
            >>> power_on_cycles >= 0
            True

        """

        return await self.read_int(address=5, offset=0, length=4)

    async def write_power_on_cycles(self, times: int):
        """Write the number of power on cycles to the EEPROM

        Args:

            times:
                The number of power on cycles that should be stored in the
                EEPROM

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Write and read the number of power on cycles of STU 1

            >>> async def write_read_power_on_cycles(times):
            ...     async with Connection() as stu:
            ...         await stu.eeprom.write_power_on_cycles(times)
            ...         return await stu.eeprom.read_power_on_cycles()
            >>> run(write_read_power_on_cycles(0))
            0

        """

        await self.write_int(address=5, offset=0, length=4, value=times)

    async def read_power_off_cycles(self) -> int:
        """Retrieve the number of power off cycles from the EEPROM

        Returns:

            The number of power off cycles of the node

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Read the number of power off cycles of STU 1

            >>> async def read_power_off_cycles():
            ...     async with Connection() as stu:
            ...         return await stu.eeprom.read_power_off_cycles()
            >>> power_off_cycles = run(read_power_off_cycles())
            >>> power_off_cycles >= 0
            True

        """

        return await self.read_int(address=5, offset=4, length=4)

    async def write_power_off_cycles(self, times: int):
        """Write the number of power off cycles to the EEPROM

        Args:

            times:
                The number of power off cycles that should be stored in the
                EEPROM

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Write and read the number of power off cycles of STU 1

            >>> async def write_read_power_off_cycles(times):
            ...     async with Connection() as stu:
            ...         await stu.eeprom.write_power_off_cycles(times)
            ...         return await stu.eeprom.read_power_off_cycles()
            >>> run(write_read_power_off_cycles(0))
            0

        """

        await self.write_int(address=5, offset=4, length=4, value=times)

    async def read_operating_time(self) -> int:
        """Retrieve the operating time from the EEPROM

        Returns:

            The operating time of the node in seconds

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Read the operating time of STU 1

            >>> async def read_operating_time():
            ...     async with Connection() as stu:
            ...         return await stu.eeprom.read_operating_time()
            >>> operating_time = run(read_operating_time())
            >>> operating_time >= 0
            True

        """

        return await self.read_int(address=5, offset=8, length=4)

    async def write_operating_time(self, seconds: int):
        """Write operating time to the EEPROM

        Args:

            seconds:
                The operating time in seconds that should be stored in the
                EEPROM

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Write and read the operating time of STU 1

            >>> async def write_read_operating_time(seconds):
            ...     async with Connection() as stu:
            ...         await stu.eeprom.write_operating_time(seconds)
            ...         return await stu.eeprom.read_operating_time()
            >>> operating_time = run(write_read_operating_time(10))
            >>> 10 <= operating_time <= 11
            True

        """

        await self.write_int(address=5, offset=8, length=4, value=seconds)

    async def read_under_voltage_counter(self) -> int:
        """Retrieve the under voltage counter value from the EEPROM

        Returns:

            The number of times the voltage was too low for the node

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Read the under voltage counter of STU 1

            >>> async def read_under_voltage_counter():
            ...     async with Connection() as stu:
            ...         return await stu.eeprom.read_under_voltage_counter()
            >>> under_voltage_counter = run(read_under_voltage_counter())
            >>> under_voltage_counter >= 0
            True

        """

        return await self.read_int(address=5, offset=12, length=4)

    async def write_under_voltage_counter(self, times: int):
        """Write the under voltage counter value to the EEPROM

        Args:

            times:
                The number of times the voltage was too low

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Write and read the under voltage counter of STU 1

            >>> async def write_read_under_voltage_counter(times):
            ...     async with Connection() as stu:
            ...         await stu.eeprom.write_under_voltage_counter(times)
            ...         return await stu.eeprom.read_under_voltage_counter()
            >>> run(write_read_under_voltage_counter(0))
            0

        """

        await self.write_int(address=5, offset=12, length=4, value=times)

    async def read_watchdog_reset_counter(self) -> int:
        """Retrieve the watchdog reset counter value from the EEPROM

        Returns:

            The watchdog reset counter value of the node

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Read the watchdog reset counter of STU 1

            >>> async def read_watchdog_reset_counter():
            ...     async with Connection() as stu:
            ...         return await stu.eeprom.read_watchdog_reset_counter()
            >>> watchdog_reset_counter = run(read_watchdog_reset_counter())
            >>> watchdog_reset_counter >= 0
            True

        """

        return await self.read_int(address=5, offset=16, length=4)

    async def write_watchdog_reset_counter(self, times: int) -> None:
        """Write the watchdog reset counter value to the EEPROM

        Args:

            times:
                The value of the watchdog reset counter for the specified node

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Write and read the watchdog reset counter of STU 1

            >>> async def write_read_watchdog_reset_counter(times):
            ...     async with Connection() as stu:
            ...         await stu.eeprom.write_watchdog_reset_counter(times)
            ...         return await stu.eeprom.read_watchdog_reset_counter()
            >>> run(write_read_watchdog_reset_counter(0))
            0

        """

        await self.write_int(address=5, offset=16, length=4, value=times)

    async def read_production_date(self) -> date:
        """Retrieve the production date from the EEPROM

        Returns:

            The production date of the specified node

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Read the production date of STU 1

            >>> async def read_production_date():
            ...     async with Connection() as stu:
            ...         return await stu.eeprom.read_production_date()
            >>> production_date = run(read_production_date())
            >>> isinstance(production_date, date)
            True

        """

        date_values = await self.read_text(address=5, offset=20, length=8)
        return date(
            year=int(date_values[0:4]),
            month=int(date_values[4:6]),
            day=int(date_values[6:8]),
        )

    # pylint: disable=redefined-outer-name

    async def write_production_date(self, date: date | str) -> None:
        """Write the production date to the EEPROM

        Args:

            date:
                The production date of the specified node

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Write and read the production date of STU 1

            >>> async def write_read_production_date(date):
            ...     async with Connection() as stu:
            ...         await stu.eeprom.write_production_date(date=date)
            ...         return await stu.eeprom.read_production_date()

            >>> production_date = date(year=2020, month=10, day=5)
            >>> str(run(write_read_production_date(production_date)))
            '2020-10-05'

            >>> production_date = '2000-01-05'
            >>> str(run(write_read_production_date(production_date)))
            '2000-01-05'

        """

        if isinstance(date, str):
            # The identifier `date` refers to the variable `date` in the
            # current scope
            import datetime  # pylint: disable=import-outside-toplevel

            try:
                date = datetime.date.fromisoformat(date)
            except ValueError as error:
                raise ValueError(
                    f"Invalid value for date argument: “{date}”"
                ) from error

        await self.write_text(
            address=5,
            offset=20,
            length=8,
            text=str(date).replace("-", ""),
        )

    # pylint: enable=redefined-outer-name

    async def read_batch_number(self) -> int:
        """Retrieve the batch number from the EEPROM

        Returns:

            The batch number of the node

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Read the batch number of STU 1

            >>> async def read_batch_number():
            ...     async with Connection() as stu:
            ...         return await stu.eeprom.read_batch_number()
            >>> batch_number = run(read_batch_number())
            >>> isinstance(batch_number, int)
            True

        """

        return await self.read_int(address=5, offset=28, length=4)

    async def write_batch_number(self, number: int) -> None:
        """Write the batch number to the EEPROM

        Args:

            number:
                The batch number of the node

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Write and read the batch number of STU 1

            >>> async def write_read_batch_number(number):
            ...     async with Connection() as stu:
            ...         await stu.eeprom.write_batch_number(number)
            ...         return await stu.eeprom.read_batch_number()
            >>> run(write_read_batch_number(1337))
            1337

        """

        await self.write_int(address=5, offset=28, length=4, value=number)


# pylint: enable=too-many-public-methods

# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



"""ICOn command line tool

See: https://mytoolit.github.io/ICOtronic/#icon-cli-tool

for more information
"""

# -- Imports ------------------------------------------------------------------

from argparse import Namespace
from asyncio import run, sleep
from functools import partial
from logging import basicConfig, getLogger
from pathlib import Path
from sys import stderr
from time import perf_counter_ns, process_time_ns, time

from can.interfaces.pcan import PcanError
from tqdm import tqdm

from icotronic.can import Connection
from icotronic.can.adc import ADCConfiguration
from icotronic.can.error import CANConnectionError, UnsupportedFeatureException
from icotronic.can.node.stu import SensorNodeInfo
from icotronic.can.node.sth import STH
from icotronic.can.streaming import StreamingTimeoutError
from icotronic.cmdline.parse import create_icon_parser
from icotronic.config import ConfigurationUtility, settings
from icotronic.measurement import convert_raw_to_g
from icotronic.measurement import Storage
from icotronic.measurement.sensor import SensorConfiguration

# -- Functions ----------------------------------------------------------------


async def get_acceleration_sensor_range_in_g(sth: STH) -> float:
    """Read sensor range of acceleration sensor

    Args:

        sth:
            The STH object used to read the sensor range

    Returns:

        The sensor range of the acceleration sensor, or the default range of
        200 (± 100 g) sensor, if there was a problem reading the sensor range

    """

    sensor_range = 200

    try:
        sensor_range = await sth.get_acceleration_sensor_range_in_g()
        if sensor_range < 1:
            print(
                f"Warning: Sensor range “{sensor_range}” below 1 g — Using "
                "range 200 instead (± 100 g sensor)",
                file=stderr,
            )
            sensor_range = 200
    except ValueError:
        print(
            "Warning: Unable to determine sensor range from "
            "EEPROM value — Assuming ± 100 g sensor",
            file=stderr,
        )

    return sensor_range


def command_config() -> None:
    """Open configuration file"""

    ConfigurationUtility.open_user_config()


# pylint: disable=too-many-locals


async def command_dataloss(arguments: Namespace) -> None:
    """Check data loss at different sample rates

    Args:

        arguments:
            The given command line arguments

    """

    identifier = arguments.identifier
    logger = getLogger()

    async with Connection() as stu:
        logger.info("Connecting to “%s”", identifier)
        async with stu.connect_sensor_node(identifier, STH) as sth:
            assert isinstance(sth, STH)
            logger.info("Connected to “%s”", identifier)

            sensor_range = await get_acceleration_sensor_range_in_g(sth)
            conversion_to_g = partial(convert_raw_to_g, max_value=sensor_range)

            measurement_time_s = 10

            sensor_config = SensorConfiguration(first=1)

            for oversampling_rate in (
                2**exponent for exponent in range(6, 10)
            ):
                logger.info("Oversampling rate: %s", oversampling_rate)
                adc_config = ADCConfiguration(
                    prescaler=2,
                    acquisition_time=8,
                    oversampling_rate=oversampling_rate,
                )
                await sth.set_adc_configuration(**adc_config)
                sample_rate = adc_config.sample_rate()
                logger.info("Sample rate: %s Hz", sample_rate)

                filepath = Path(f"Measurement {sample_rate} Hz.hdf5")
                with Storage(
                    filepath.resolve(), sensor_config.streaming_configuration()
                ) as storage:
                    storage.write_sensor_range(sensor_range)
                    storage.write_sample_rate(adc_config)

                    progress = tqdm(
                        total=int(sample_rate * measurement_time_s),
                        desc="Read sensor data",
                        unit=" values",
                        leave=False,
                        disable=None,
                    )

                    start_time = time()
                    try:
                        async with sth.open_data_stream(
                            sensor_config.streaming_configuration()
                        ) as stream:
                            async for data, _ in stream:
                                storage.add_streaming_data(
                                    data.apply(conversion_to_g)
                                )
                                progress.update(3)
                                if time() - start_time >= measurement_time_s:
                                    break
                    except PcanError as error:
                        print(
                            f"Unable to collect streaming data: {error}",
                            file=stderr,
                        )

                    progress.close()
                print(f"Stored measurement data in “{filepath}”")


# pylint: enable=too-many-locals


async def command_list(
    arguments: Namespace,  # pylint: disable=unused-argument
) -> None:
    """Print a list of available sensor nodes

    Args:

        arguments:
            The given command line arguments

    """

    async with Connection() as stu:
        timeout = time() + 5
        sensor_nodes: list[SensorNodeInfo] = []
        sensor_nodes_before: list[SensorNodeInfo] = []

        # - First request for sensor nodes will produce empty list
        # - Subsequent retries should provide all available sensor nodes
        # - We wait until the number of sensor nodes is larger than 1 and
        #   has not changed between one iteration or the timeout is reached
        while (
            len(sensor_nodes) <= 0
            and time() < timeout
            or len(sensor_nodes) != len(sensor_nodes_before)
        ):
            sensor_nodes_before = list(sensor_nodes)
            sensor_nodes = await stu.get_sensor_nodes()
            await sleep(0.5)

        for node in sensor_nodes:
            print(node)


# pylint: disable=too-many-locals


async def command_measure(arguments: Namespace) -> None:
    """Open measurement stream and store data

    Args:

        arguments:
            The given command line arguments

    """

    logger = getLogger()

    identifier = arguments.identifier
    measurement_time_s = arguments.time

    async with Connection() as stu:
        async with stu.connect_sensor_node(identifier, STH) as sth:
            assert isinstance(sth, STH)

            adc_config = ADCConfiguration(
                reference_voltage=arguments.voltage_reference,
                prescaler=arguments.prescaler,
                acquisition_time=arguments.acquisition,
                oversampling_rate=arguments.oversampling,
            )
            await sth.set_adc_configuration(**adc_config)
            print(f"Sample Rate: {adc_config.sample_rate():.2f} Hz")

            user_sensor_config = SensorConfiguration(
                first=arguments.first_channel,
                second=arguments.second_channel,
                third=arguments.third_channel,
            )

            if user_sensor_config.requires_channel_configuration_support():
                try:
                    await sth.set_sensor_configuration(user_sensor_config)
                except UnsupportedFeatureException as exception:
                    raise UnsupportedFeatureException(
                        f"Sensor channel configuration “{user_sensor_config}”"
                        f" is not supported by the sensor node “{identifier}”"
                    ) from exception

            sensor_range = await get_acceleration_sensor_range_in_g(sth)
            conversion_to_g = partial(convert_raw_to_g, max_value=sensor_range)
            filepath = settings.get_output_filepath()

            with Storage(
                filepath, user_sensor_config.streaming_configuration()
            ) as storage:
                storage.write_sensor_range(sensor_range)
                storage.write_sample_rate(adc_config)

                streaming_config = user_sensor_config.streaming_configuration()
                logger.info("Streaming Configuration: %s", streaming_config)
                values_per_message = streaming_config.data_length()

                progress = tqdm(
                    total=round(
                        adc_config.sample_rate() * measurement_time_s, 0
                    ),
                    desc="Read sensor data",
                    unit=" values",
                    leave=False,
                    disable=None,
                )

                try:
                    async with sth.open_data_stream(
                        streaming_config
                    ) as stream:
                        start_time = time()
                        async for data, _ in stream:
                            storage.add_streaming_data(
                                data.apply(conversion_to_g)
                            )
                            progress.update(values_per_message)
                            if time() - start_time >= measurement_time_s:
                                break
                except KeyboardInterrupt:
                    pass
                finally:
                    progress.close()
                    print(f"Data Loss: {storage.dataloss() * 100} %")
                    print(f"Filepath: {filepath}")


# pylint: enable=too-many-locals


async def command_rename(arguments: Namespace) -> None:
    """Rename a sensor node

    Args:

        arguments:
            The given command line arguments

    """

    identifier = arguments.identifier
    name = arguments.name

    async with Connection() as stu:
        async with stu.connect_sensor_node(identifier) as sensor_node:
            old_name = await sensor_node.get_name()
            mac_address = await sensor_node.get_mac_address()

            await sensor_node.set_name(name)
            name = await sensor_node.get_name()
            print(
                f"Renamed sensor node “{old_name}” with MAC "
                f"address “{mac_address}” to “{name}”"
            )


async def command_stu(arguments: Namespace) -> None:
    """Run specific commands regarding stationary transceiver unit

    Args:

        arguments:
            The given command line arguments

    """

    subcommand = arguments.stu_subcommand

    async with Connection() as stu:
        if subcommand == "ota":
            # The coroutine below activates the advertisement required for the
            # Over The Air (OTA) firmware update.
            #
            # - The `deactivate_bluetooth` command called when the execution
            #   leaves the `async with` block seems to not turn off the
            #   advertisement for the STU.
            # - Even a **hard STU reset does not turn off the advertisement**.
            # - One way to turn off the advertisement seems to be to initiate a
            #   connection with a sensor node.
            await stu.activate_bluetooth()
        elif subcommand == "mac":
            print(await stu.get_mac_address())
        elif subcommand == "reset":
            await stu.reset()
        else:
            raise ValueError(f"Unknown STU subcommand “{subcommand}”")


def main():
    """ICOtronic command line tool"""

    parser = create_icon_parser()
    arguments = parser.parse_args()
    try:
        if arguments.subcommand == "measure":
            SensorConfiguration(
                first=arguments.first_channel,
                second=arguments.second_channel,
                third=arguments.third_channel,
            ).check()
    except ValueError as error:
        parser.prog = f"{parser.prog} {arguments.subcommand}"
        parser.error(str(error))

    basicConfig(
        level=arguments.log.upper(),
        style="{",
        format="{asctime} {levelname:7} {message}",
    )

    logger = getLogger()
    logger.info("CLI Arguments: %s", arguments)

    if arguments.subcommand == "config":
        command_config()
    else:
        command_to_coroutine = {
            "dataloss": command_dataloss,
            "list": command_list,
            "measure": command_measure,
            "rename": command_rename,
            "stu": command_stu,
        }

        try:
            perf_start, cpu_start = perf_counter_ns(), process_time_ns()
            run(command_to_coroutine[arguments.subcommand](arguments))
            perf_end, cpu_end = perf_counter_ns(), process_time_ns()
            run_time_command = perf_end - perf_start
            cpu_time_command = cpu_end - cpu_start
            cpu_usage = cpu_time_command / run_time_command * 100
            logger.info(
                "Ran command “%s” in %.2f seconds (CPU time: %.2f seconds, "
                "CPU Usage: %.2f %%)",
                arguments.subcommand,
                run_time_command / 10**9,
                cpu_time_command / 10**9,
                cpu_usage,
            )
        except (
            CANConnectionError,
            TimeoutError,
            UnsupportedFeatureException,
            ValueError,
        ) as error:
            print(error, file=stderr)
        except StreamingTimeoutError as error:
            print(f"Quitting Measurement: {error}")
        except KeyboardInterrupt:
            pass


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    main()



"""Support code for sensors and sensor configuration"""

# -- Imports ------------------------------------------------------------------

from collections.abc import Iterable, Iterator, Mapping
from enum import auto, Enum
from statistics import mean
from typing import NamedTuple

from icotronic.can.streaming import StreamingConfiguration

# -- Classes ------------------------------------------------------------------


class SensorConfiguration(Mapping):
    """Used to store the configuration of the three sensor channels

    Args:

        first:
            The sensor number for the first measurement channel

        second:
            The sensor number for the second measurement channel

        third:
            The sensor number for the third measurement channel


    Examples:

        Create an example sensor configuration

        >>> SensorConfiguration(first=0, second=1, third=2)
        M1: None, M2: S1, M3: S2

        Initializing a sensor configuration with incorrect values will fail

        >>> SensorConfiguration(first=256, second=1, third=2)
        Traceback (most recent call last):
        ...
        ValueError: Incorrect value for first channel: “256”

        >>> SensorConfiguration(first=0, second=1, third=-1)
        Traceback (most recent call last):
        ...
        ValueError: Incorrect value for third channel: “-1”

    """

    def __init__(self, first: int = 0, second: int = 0, third: int = 0):

        self.attributes = {
            "first": first,
            "second": second,
            "third": third,
        }

        for name, channel in self.attributes.items():
            if channel < 0 or channel > 255:
                raise ValueError(
                    f"Incorrect value for {name} channel: “{channel}”"
                )

    def __getitem__(self, item: str) -> int:
        """Return values of the mapping provided by this class

        Note:

            This method allow access to the object via the splat operators (*,
            **)

        Args:

            item:
                The attribute for which we want to retrieve the value

        Returns:

            The value of the attribute

        Examples:

            Create an “empty” example sensor configuration

            >>> dict(**SensorConfiguration()) # doctest:+NORMALIZE_WHITESPACE
            {'first': 0,
             'second': 0,
             'third': 0}

            Create an example sensor configuration using init values

            >>> dict(**SensorConfiguration(first=1, second=2, third=3)
            ...     ) # doctest:+NORMALIZE_WHITESPACE
            {'first': 1,
             'second': 2,
             'third': 3}

        """

        return self.attributes[item]

    def __iter__(self) -> Iterator:
        """Return an iterator over the mapping provided by this class

        Note:

            This method allow access to the object via the splat operators (*,
            **)

        Returns:

            The names of the “important” properties of the sensor
            configuration:

            - first
            - second
            - third

        Examples:

            Get the important sensor config attributes

            >>> for attribute in SensorConfiguration():
            ...     print(attribute)
            first
            second
            third

        """

        return iter(self.attributes)

    def __len__(self) -> int:
        """Return the length of the mapping provided by this class

        Note:

            This method allow access to the object via the splat operators (*,
            **)

        Returns:

            The number of “important” properties of the sensor configuration:

            - first
            - second
            - third

        Examples:

            Get the static length of some example sensor configurations

            >>> len(SensorConfiguration())
            3

            >>> len(SensorConfiguration(second=10))
            3

        """

        return len(self.attributes)

    def __str__(self) -> str:
        """The string representation of the sensor configuration

        Returns:

            A textual representation of the sensor configuration

        Examples:


            Get the string representation of some example sensor configs

            >>> str(SensorConfiguration(first=1, second=3, third=2))
            'M1: S1, M2: S3, M3: S2'

            >>> str(SensorConfiguration())
            ''

            >>> str(SensorConfiguration(second=1))
            'M2: S1'

        """

        return ", ".join((
            f"M{sensor}: S{value}"
            for sensor, value in enumerate(self.attributes.values(), start=1)
            if value != 0
        ))

    def __repr__(self) -> str:
        """The textual representation of the sensor configuration

        Returns:

            A textual representation of the sensor configuration

        Examples:

            Get the textual representation of some example sensor configs

            >>> repr(SensorConfiguration(first=1, second=3, third=2))
            'M1: S1, M2: S3, M3: S2'

            >>> repr(SensorConfiguration())
            'M1: None, M2: None, M3: None'

        """

        return ", ".join((
            f"M{sensor}: {f'S{value}' if value != 0 else 'None'}"
            for sensor, value in enumerate(self.attributes.values(), start=1)
        ))

    @property
    def first(self) -> int:
        """Get the sensor for the first channel

        Returns:

            The sensor number of the first channel

        Examples:

            Get the sensor number for the first channel of a sensor config

            >>> SensorConfiguration(first=1, second=3, third=2).first
            1

        """

        first = self.attributes["first"]

        return 0 if first is None else first

    @property
    def second(self) -> int:
        """Get the sensor for the second channel

        Returns:

            The sensor number of the second channel


        Examples:

            Get the sensor number for the second channel of a sensor config

            >>> SensorConfiguration(first=1, second=3, third=2).second
            3

        """

        second = self.attributes["second"]

        return 0 if second is None else second

    @property
    def third(self) -> int:
        """Get the sensor for the third channel

        Returns:

            The sensor number of the third channel

        Examples:

            Get the sensor number for the third channel of a sensor config

            >>> SensorConfiguration(first=1, second=3, third=2).third
            2

        """

        third = self.attributes["third"]

        return 0 if third is None else third

    def disable_channel(
        self, first: bool = False, second: bool = False, third: bool = False
    ) -> None:
        """Disable certain (measurement) channels

        Args:

            first:
                Specifies if the first measurement channel should be disabled
                or not

            second:
                Specifies if the second measurement channel should be disabled
                or not

            third:
                Specifies if the third measurement channel should be disabled
                or not

        """

        if first:
            self.attributes["first"] = 0
        if second:
            self.attributes["second"] = 0
        if third:
            self.attributes["third"] = 0

    def requires_channel_configuration_support(self) -> bool:
        """Check if the sensor configuration requires channel config support

        Returns:

            - ``True``, if the configuration requires hardware that has
              support for changing the channel configuration
            - ``False``, otherwise

        Examples:

            Check if example sensor configs require channel config support

            >>> SensorConfiguration(first=1, second=3, third=2
            ...     ).requires_channel_configuration_support()
            True

            >>> SensorConfiguration(first=1, second=0, third=1
            ...     ).requires_channel_configuration_support()
            True

            >>> SensorConfiguration(first=1, second=2, third=3
            ...     ).requires_channel_configuration_support()
            False

            >>> SensorConfiguration().requires_channel_configuration_support()
            False

        """

        for measurement_channel, sensor_channel in enumerate(
            self.attributes.values(), start=1
        ):
            if sensor_channel not in {0, measurement_channel}:
                return True
        return False

    def empty(self) -> bool:
        """Check if the sensor configuration is empty

        In an empty sensor configuration all of the channels are disabled.

        Returns:

            ``True``, if all channels are disabled, ``False`` otherwise

        Examples:

            Check if some example configurations are empty or not

            >>> SensorConfiguration(first=3).empty()
            False
            >>> SensorConfiguration().empty()
            True
            >>> SensorConfiguration(third=0).empty()
            True

        """

        return self.first == 0 and self.second == 0 and self.third == 0

    def check(self):
        """Check that at least one measurement channel is enabled

        Raises:

            ValueError:
                if none of the measurement channels is enabled

        Examples:

            >>> SensorConfiguration(second=1).check()
            >>> SensorConfiguration().check()
            Traceback (most recent call last):
                ...
            ValueError: At least one measurement channel has to be enabled

        """

        if self.empty():
            raise ValueError(
                "At least one measurement channel has to be enabled"
            )

    def streaming_configuration(self) -> StreamingConfiguration:
        """Get a streaming configuration that represents this config

        Returns:

            A stream configuration where

            - every channel that is enabled in the sensor configuration is
              enabled, and
            - every channel that is disables in the sensor configuration is
              disabled.

        Examples:

            Get the streaming configuration for some example channel configs

            >>> SensorConfiguration(second=1).streaming_configuration()
            Channel 1 disabled, Channel 2 enabled, Channel 3 disabled

            >>> SensorConfiguration(first=10, third=2
            ...                    ).streaming_configuration()
            Channel 1 enabled, Channel 2 disabled, Channel 3 enabled

        """

        return StreamingConfiguration(**{
            channel: bool(value) for channel, value in self.attributes.items()
        })


class SensorType(Enum):
    """Possible sensor types"""

    BROKEN = auto()
    """Broken sensor"""

    ACCELERATION = auto()
    """Acceleration sensor"""

    TEMPERATURE = auto()
    """Temperature sensor"""

    PIEZO = auto()
    """Piezoelectric sensor"""


class Sensor(NamedTuple):
    """Store information about a sensor"""

    type: SensorType
    """Sensor type"""

    mean: float
    """Mean value of raw ADC sensor data"""

    def __repr__(self) -> str:
        """Return a string representation of the sensor

        Returns:

            A string that describes the sensor

        Examples:

            Get the textual representation of some example sensor objects

            >>> Sensor(SensorType.BROKEN, mean=123)
            Broken Sensor (Mean: 123)

            >>> Sensor(SensorType.TEMPERATURE, mean=123)
            Temperature Sensor

        """

        representation = f"{self.type.name.capitalize()} Sensor"
        if self.type == SensorType.BROKEN:
            representation += f" (Mean: {self.mean})"

        return representation

    def works(self) -> bool:
        """Check if the sensor is working or not

        Returns:

            ``True`` if the sensor works, ``False`` otherwise

        """

        return not self.type == SensorType.BROKEN


# -- Functions ----------------------------------------------------------------


def guess_sensor(values: Iterable[int]) -> Sensor:
    """Guess the sensor type from raw 16 bit ADC values

    Args:

        values:
            Multiple raw 16 bit ADC measurement values

    Returns:

        An object representing the guessed sensor type

    Examples:

        Guess sensor types based on raw ADC values

        >>> guess_sensor([38024, 38000, 37950])
        Piezo Sensor

        >>> guess_sensor([32500, 32571, 32499])
        Acceleration Sensor

        >>> guess_sensor([10650, 10500, 10780])
        Temperature Sensor

        >>> guess_sensor([123, 10, 50])
        Broken Sensor (Mean: 61)

    """

    mean_raw = mean(values)
    half = 2**15

    tolerance_acceleration = 1000
    min_acceleration = half - tolerance_acceleration
    max_acceleration = half + tolerance_acceleration

    if 10000 <= mean_raw <= 11000:
        return Sensor(SensorType.TEMPERATURE, mean_raw)
    if min_acceleration <= mean_raw <= max_acceleration:
        return Sensor(SensorType.ACCELERATION, mean_raw)
    if 37500 <= mean_raw <= 38500:
        return Sensor(SensorType.PIEZO, mean_raw)

    return Sensor(SensorType.BROKEN, mean_raw)


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



"""Define block and block commands for the ICOtronic system

See also: https://mytoolit.github.io/Documentation/#blocks
"""

# -- Classes ------------------------------------------------------------------


class UnknownBlockCommandError(Exception):
    """Raised if the block command is undefined/unknown"""


class UnknownBlockError(Exception):
    """Raised if the block is undefined/unknown"""


# pylint: disable=too-few-public-methods


class CANIdPart:
    """Stores data for a part of the CAN identifier

    Args:

        number:
            The value of the CAN identifier part

        name:
            The name of the CAN identifier part

        description:
            A short description of the CAN identifier part

    """

    def __init__(self, number: int, name: str, description: str) -> None:

        self.number = number
        self.name = name
        self.description = description

    def __repr__(self) -> str:
        """Get the string representation of the CAN identifier part

        Examples:

            Get the textual representation of example CAN identifier info

            >>> CANIdPart(0x01, "Reset", "System Command Reset")
            Reset (0x01): System Command Reset

            >>> CANIdPart(0x80, "Energy", "Statistical Data Command Energy")
            Energy (0x80): Statistical Data Command Energy

        Returns:

            A string representing the CAN identifier part

        """

        return f"{self.name} (0x{self.number:02X}): {self.description}"


class Block(CANIdPart):
    """Store information about a block including its block commands

    Args:

        number:
            The value of the CAN identifier part

        name:
            The name of the CAN identifier part

        description:
            A short description of the CAN identifier part

        block_commands:
            The list of block commands contained in this block

    """

    def __init__(
        self,
        number: int,
        name: str,
        description: str,
        block_commands: list[CANIdPart],
    ) -> None:

        super().__init__(number, name, description)
        self.block_commands: dict[int | str, CANIdPart] = {}
        for block_command in block_commands:
            self.block_commands[block_command.number] = block_command
            self.block_commands[block_command.name] = block_command

    def __getitem__(self, block_command: int | str) -> CANIdPart:
        """Get the block command with the specified number or name

        Args:

            block_command:
                The name or number of the block command

        Returns:

            The requested block command

        Raises:

            UnknownBlockCommandError
                if the command block is not part of the block

        Examples:

            Get block command by number

            >>> block = Block(
            ...     0x00,
            ...     "System",
            ...     "Command Block System",
            ...     [
            ...         CANIdPart(0x00, "Verboten", "System Command Verboten"),
            ...         CANIdPart(0x01, "Reset", "System Command Reset"),
            ...     ],
            ... )
            >>> block[0]
            Verboten (0x00): System Command Verboten
            >>> block[0] == block["Verboten"]
            True

            Accessing a non existent block command should fail

            >>> block[-1] # doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
                ...
            UnknownBlockCommandError: Unknown block command: “-1”

        """

        try:
            return self.block_commands[block_command]
        except KeyError as error:
            raise UnknownBlockCommandError(
                f"Unknown block command: “{block_command}”"
            ) from error


class Blocks:
    """Store information about available blocks

    Args:

        block_info:
            A list containing each block

    """

    def __init__(self, block_info: list[Block]):

        self.blocks: dict[int | str, Block] = {}
        for block in block_info:
            self.blocks[block.number] = block
            self.blocks[block.name] = block

    def __getitem__(self, block: int | str) -> Block:
        """Get the block with the specified number or name

        Args:

            block:
                The name or number of the block

        Returns:

            The requested block

        Raises:

            UnknownBlockCommandError:
                if the command block is not part of the block

        Examples:

            Create a Blocks object

            >>> blocks = Blocks([
            ...     Block(
            ...         0x00,
            ...         "System",
            ...         "Command Block System",
            ...         [
            ...             CANIdPart(0x00, "Verboten",
            ...                       "System Command Verboten"),
            ...             CANIdPart(0x01, "Reset", "System Command Reset"),
            ...         ],
            ... )])

            Retrieve an existing block

            >>> blocks["System"]
            System (0x00): Command Block System

            Retrieving an non-existing block will fail

            >>> blocks["Does Not Exist"] # doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
                ...
            UnknownBlockError: Unknown block: “Does Not Exist”

        """

        try:
            return self.blocks[block]
        except KeyError as error:
            raise UnknownBlockError(f"Unknown block: “{block}”") from error

    def __repr__(self) -> str:
        """Get the string representation of the blocks

        Returns:

            The string representation of the stored blocks

        Examples:

            Get the textual representation of a Blocks object

            >>> Blocks([
            ...     Block(
            ...         0x3F,
            ...         "Test",
            ...         "Command Block Test",
            ...         [
            ...             CANIdPart(0x01, "Signal", "Test Command Signal"),
            ...             CANIdPart(0x69, "Pfeifferl",
            ...                       "Test Command Pfeifferl"),
            ...         ]
            ...     ),
            ...     Block(0x3E, "Product Data",
            ...           "Command Block Product Data", [])
            ... ]) # doctest: +NORMALIZE_WHITESPACE
            [Test (0x3F): Command Block Test,
             Product Data (0x3E): Command Block Product Data]

        """

        return (
            "["
            + ", ".join([
                repr(block)
                for name, block in self.blocks.items()
                if isinstance(name, str)
            ])
            + "]"
        )


# pylint: enable=too-few-public-methods

blocks = Blocks([
    Block(
        0x00,
        "System",
        "Command Block System",
        [
            CANIdPart(0x00, "Verboten", "System Command Verboten"),
            CANIdPart(0x01, "Reset", "System Command Reset"),
            CANIdPart(0x02, "Get/Set State", "System Command Get/Set State"),
            CANIdPart(0x03, "Mode", "System Command Mode"),
            CANIdPart(0x04, "Alarm", "System Command Alarm"),
            CANIdPart(0x05, "Node Status", "System Command Status Word0"),
            CANIdPart(0x06, "Error Status", "System Command Status Word1"),
            CANIdPart(0x07, "StatusWord2", "System Command Status Word2"),
            CANIdPart(0x08, "StatusWord3", "System Command Status Word3"),
            CANIdPart(0x09, "Test", "System Command Test"),
            CANIdPart(0x0A, "Log", "System Command Log"),
            CANIdPart(0x0B, "Bluetooth", "System Command BlueTooth"),
            CANIdPart(0x0C, "Routing", "System Command Routing"),
        ],
    ),
    Block(
        0x04,
        "Streaming",
        "Command Block Streaming",
        [
            CANIdPart(0x00, "Data", "Streaming Command Data"),
            CANIdPart(0x01, "Temperature", "Streaming Command Temperature"),
            CANIdPart(0x20, "Voltage", "Streaming Command Voltage"),
            CANIdPart(0x40, "Current", "Streaming Command Current"),
        ],
    ),
    Block(
        0x08,
        "StatisticalData",
        "Command Block Statistical Data",
        [
            CANIdPart(
                0x00,
                "PowOnOff",
                "Statistical Data Command PowerOn/Off Counter",
            ),
            CANIdPart(
                0x01,
                "OperatingTime",
                "Statistical Data Command Operating Time",
            ),
            CANIdPart(
                0x02, "UVC", "Statistical Data Command Under-Voltage Counter"
            ),
            CANIdPart(
                0x03, "WDog", "Statistical Data Command Watchdog Counter"
            ),
            CANIdPart(
                0x04,
                "ProductionDate",
                "Statistical Data Command Production Date",
            ),
            CANIdPart(
                0x40,
                "MeasurementInterval",
                "Statistical Data Command Measurement Interval",
            ),
            CANIdPart(
                0x41,
                "QuantityInterval",
                "Statistical Data Command Quantity Interval",
            ),
            CANIdPart(0x80, "Energy", "Statistical Data Command Energy"),
        ],
    ),
    Block(
        0x28,
        "Configuration",
        "Command Block Configuration",
        [
            CANIdPart(
                0x00,
                "Get/Set ADC Configuration",
                "Configuration Command Acceleration Configuration",
            ),
            CANIdPart(
                0x01,
                "Channel Configuration",
                "Configuration Command Channel Configuration",
            ),
            CANIdPart(
                0x20, "Voltage", "Configuration Command Voltage Configuration"
            ),
            CANIdPart(
                0x40, "Current", "Configuration Command Current Configuration"
            ),
            CANIdPart(
                0x60,
                "CalibrationFactorK",
                "Configuration Command Calibration Factor K",
            ),
            CANIdPart(
                0x61,
                "CalibrationFactorD",
                "Configuration Command Calibration Factor D",
            ),
            CANIdPart(
                0x62,
                "Calibration Measurement",
                "Configuration Command Calibration Measurement",
            ),
            CANIdPart(0x80, "Alarm", "Configuration Command Alarm"),
            CANIdPart(0xC0, "HMI", "Configuration Command HMI"),
        ],
    ),
    Block(
        0x3D,
        "EEPROM",
        "Command Block EEPROM",
        [
            CANIdPart(0x00, "Read", "EEPROM Command Read"),
            CANIdPart(0x01, "Write", "EEPROM Command Write"),
            CANIdPart(
                0x20,
                "Read Write Request Counter",
                "EEPROM Command Write Request Counter",
            ),
        ],
    ),
    Block(
        0x3E,
        "Product Data",
        "Command Block Product Data",
        [
            CANIdPart(0x00, "GTIN", "Product Data Command GTIN"),
            CANIdPart(
                0x01,
                "Hardware Version",
                "Product Data Command Hardware Version",
            ),
            CANIdPart(
                0x02,
                "Firmware Version",
                "Product Data Command Firmware Version",
            ),
            CANIdPart(
                0x03, "Release Name", "Product Data Command Release Name"
            ),
            CANIdPart(
                0x04, "Serial Number 1", "Product Data Command Serial Number 1"
            ),
            CANIdPart(
                0x05, "Serial Number 2", "Product Data Command Serial Number 2"
            ),
            CANIdPart(
                0x06, "Serial Number 3", "Product Data Command Serial Number 3"
            ),
            CANIdPart(
                0x07, "Serial Number 4", "Product Data Command Serial Number 4"
            ),
            CANIdPart(0x08, "Product Name 1", "Product Data Command Name1"),
            CANIdPart(0x09, "Product Name 2", "Product Data Command Name2"),
            CANIdPart(0x0A, "Product Name 3", "Product Data Command Name3"),
            CANIdPart(0x0B, "Product Name 4", "Product Data Command Name4"),
            CANIdPart(0x0C, "Product Name 5", "Product Data Command Name5"),
            CANIdPart(0x0D, "Product Name 6", "Product Data Command Name6"),
            CANIdPart(0x0E, "Product Name 7", "Product Data Command Name7"),
            CANIdPart(0x0F, "Product Name 8", "Product Data Command Name8"),
            CANIdPart(0x10, "Product Name 9", "Product Data Command Name9"),
            CANIdPart(0x11, "Product Name 10", "Product Data Command Name10"),
            CANIdPart(0x12, "Product Name 11", "Product Data Command Name11"),
            CANIdPart(0x13, "Product Name 12", "Product Data Command Name12"),
            CANIdPart(0x14, "Product Name 13", "Product Data Command Name13"),
            CANIdPart(0x15, "Product Name 14", "Product Data Command Name14"),
            CANIdPart(0x16, "Product Name 15", "Product Data Command Name15"),
            CANIdPart(0x17, "Product Name 16", "Product Data Command Name16"),
            CANIdPart(
                0x18, "OEM Free Use 1", "Product Data Command Free Use 1"
            ),
            CANIdPart(
                0x19, "OEM Free Use 2", "Product Data Command Free Use 2"
            ),
            CANIdPart(
                0x1A, "OEM Free Use 3", "Product Data Command Free Use 3"
            ),
            CANIdPart(
                0x1B, "OEM Free Use 4", "Product Data Command Free Use 4"
            ),
            CANIdPart(
                0x1C, "OEM Free Use 5", "Product Data Command Free Use 5"
            ),
            CANIdPart(
                0x1D, "OEM Free Use 6", "Product Data Command Free Use 6"
            ),
            CANIdPart(
                0x1E, "OEM Free Use 7", "Product Data Command Free Use 7"
            ),
            CANIdPart(
                0x1F, "OEM Free Use 8", "Product Data Command Free Use 8"
            ),
        ],
    ),
    Block(
        0x3F,
        "Test",
        "Command Block Test",
        [
            CANIdPart(0x01, "Signal", "Test Command Signal"),
            CANIdPart(0x69, "Pfeifferl", "Test Command Pfeifferl"),
        ],
    ),
])
"""MyTooliT Protocol command blocks

See also: https://mytoolit.github.io/Documentation/#blocks
"""

# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



"""Support for streaming (measurement) data in the ICOtronic system"""

# -- Imports ------------------------------------------------------------------

from __future__ import annotations

from asyncio import Queue, wait_for
from collections.abc import AsyncIterator, Callable, Sequence
from ctypes import c_uint8, LittleEndianStructure

from can import Listener, Message

from icotronic.can.protocol.identifier import Identifier

# -- Classes ------------------------------------------------------------------


class StreamingError(Exception):
    """General exception for streaming errors"""


class StreamingTimeoutError(StreamingError):
    """Raised if no streaming data was received for a certain amount of time"""


class StreamingBufferError(StreamingError):
    """Raised if there are too many streaming messages in the buffer"""


# pylint: disable=too-few-public-methods


class StreamingConfigBits(LittleEndianStructure):
    """Store enable/disabled channels of streaming configuration"""

    _fields_ = [
        ("first", c_uint8, 1),
        ("second", c_uint8, 1),
        ("third", c_uint8, 1),
    ]


# pylint: enable=too-few-public-methods


class StreamingConfiguration:
    """Streaming configuration

    Args:

        first:
            Specifies if the first channel is enabled or not

        second:
            Specifies if the second channel is enabled or not

        third:
            Specifies if the third channel is enabled or not

    Raises:

        ValueError:
            if none of the channels is active

    Examples:

        Create some example streaming configurations

        >>> config = StreamingConfiguration()
        >>> config = StreamingConfiguration(
        ...          first=False, second=True, third=True)

        Creating streaming configurations without active channels will fail

        >>> config = StreamingConfiguration(first=False)
        Traceback (most recent call last):
           ...
        ValueError: At least one channel needs to be active

        >>> config = StreamingConfiguration(
        ...     first=False, second=False, third=False)
        Traceback (most recent call last):
           ...
        ValueError: At least one channel needs to be active

    """

    def __init__(
        self, first: bool = True, second: bool = False, third: bool = False
    ) -> None:

        if first + second + third <= 0:
            raise ValueError("At least one channel needs to be active")

        self.channels = StreamingConfigBits(
            first=first, second=second, third=third
        )

    def __repr__(self) -> str:
        """Return the string representation of the streaming configuration

        Examples:

            Get the textual representation of some example streaming configs

            >>> StreamingConfiguration()
            Channel 1 enabled, Channel 2 disabled, Channel 3 disabled

            >>> StreamingConfiguration(first=False, second=True, third=False)
            Channel 1 disabled, Channel 2 enabled, Channel 3 disabled

            >>> StreamingConfiguration(first=True, second=True, third=True)
            Channel 1 enabled, Channel 2 enabled, Channel 3 enabled

        """

        channels = self.channels

        return ", ".join([
            f"Channel {name} {'en' if status else 'dis'}abled"
            for name, status in enumerate(
                (channels.first, channels.second, channels.third), start=1
            )
        ])

    def enabled_channels(self) -> int:
        """Get the number of activated channels

        Returns:

            The number of enabled channels

        Examples:

            Get the number of enabled channels for example streaming configs

            >>> StreamingConfiguration(first=True).enabled_channels()
            1

            >>> StreamingConfiguration(first=False, second=True, third=False
            ...                       ).enabled_channels()
            1

            >>> StreamingConfiguration(first=True, second=True, third=True
            ...                       ).enabled_channels()
            3

        """

        channels = self.channels

        return channels.first + channels.second + channels.third

    def data_length(self) -> int:
        """Returns the streaming data length

        This will be either:

        - 2 (when 2 channels are active), or
        - 3 (when 1 or 3 channels are active)

        For more information, please take a look `here`_.

        .. _here: https://mytoolit.github.io/Documentation/#command-data

        Returns:

            The length of the streaming data resulting from this channel
            configuration

        Examples:

            Get the data length of example streaming configurations

            >>> StreamingConfiguration().data_length()
            3

            >>> StreamingConfiguration(
            ...     first=False, second=True, third=False).data_length()
            3

            >>> StreamingConfiguration(
            ...     first=True, second=True, third=True).data_length()
            3

            >>> StreamingConfiguration(
            ...     first=False, second=True, third=True).data_length()
            2

        """

        return 2 if self.enabled_channels() == 2 else 3

    def axes(self) -> list[str]:
        """Get the activated axes returned by this streaming configuration

        Returns:

            A list containing all activated axes in alphabetical order

        Examples:

            Get the activated axes for example streaming configurations

            >>> StreamingConfiguration(
            ...     first=False, second=True, third=True).axes()
            ['y', 'z']
            >>> StreamingConfiguration(
            ...     first=True, second=True, third=False).axes()
            ['x', 'y']

        """

        channels = self.channels
        return [
            axis
            for axis, status in zip(
                "xyz",
                (channels.first, channels.second, channels.third),
            )
            if status
        ]

    @property
    def first(self) -> bool:
        """Check the activation state of the first channel

        Returns:

        ``True``, if the first channel is enabled or ``False`` otherwise

        Examples:

            Check channel one activation status for example configs

            >>> StreamingConfiguration(first=True, second=False,
            ...                        third=False).first
            True
            >>> StreamingConfiguration(first=False, second=False,
            ...                        third=True).first
            False

        """

        return bool(self.channels.first)

    @property
    def second(self) -> bool:
        """Check the activation state of the second channel

        Returns:

            ``True``, if the second channel is enabled or ``False`` otherwise

        Examples:

            Check channel two activation status for example configs

            >>> StreamingConfiguration(
            ...     first=True, second=False, third=False).second
            False
            >>> StreamingConfiguration(
            ...     first=False, second=True, third=True).second
            True

        """

        return bool(self.channels.second)

    @property
    def third(self) -> bool:
        """Check the activation state of the third channel

        Returns:

        ``True``, if the third channel is enabled or ``False`` otherwise

        Examples:

            Check channel three activation status for example configs

            >>> StreamingConfiguration(
            ...     first=True, second=False, third=False).third
            False
            >>> StreamingConfiguration(
            ...     first=False, second=False, third=True).third
            True

        """

        return bool(self.channels.third)


# pylint: disable=too-few-public-methods


class MessageStats:
    """Store message statistics

    Args:

        retrieved:
            The number of successfully retrieved messages

    """

    def __init__(self, retrieved: int = 0, lost: int = 0):

        self.retrieved = retrieved
        self.lost = lost

    def dataloss(self) -> float:
        """Get the amount of data loss

        Returns:

            The overall amount of data loss as number between 0 (no data loss)
            and 1 (all data lost).

        Examples:

            >>> MessageStats().dataloss()
            0

            >>> MessageStats(50, 50).dataloss()
            0.5

        """

        overall = self.retrieved + self.lost

        return 0 if overall == 0 else self.lost / overall

    def reset(self) -> None:
        """Reset the amount of retrieved and lost messages to 0

        Examples:

            >>> stats = MessageStats(10, 90)
            >>> stats.dataloss()
            0.9
            >>> stats.reset()
            >>> stats.dataloss()
            0

        """

        self.retrieved = 0
        self.lost = 0


# pylint: enable=too-few-public-methods


class AsyncStreamBuffer(Listener):
    """Buffer for streaming data

    Args:

        configuration:
            A streaming configuration that specifies which of the three
            streaming channels should be enabled or not

        timeout:
            The amount of seconds between two consecutive messages, before
            a ``StreamingTimeoutError`` will be raised

        max_buffer_size:
            Maximum amount of buffered messages kept by the stream buffer.
            If this amount is exceeded, then this listener will raise a
            ``StreamingBufferError``. A large buffer indicates that the
            application is not able to keep up with the current rate of
            retrieved messages and therefore the probability of losing
            messages is quite high.

    """

    def __init__(
        self,
        configuration: StreamingConfiguration,
        timeout: float,
        max_buffer_size: int,
    ) -> None:

        # Expected identifier of received streaming messages
        self.identifier = Identifier(
            block="Streaming",
            block_command="Data",
            sender="STH 1",
            receiver="SPU 1",
            request=False,
        )
        self.queue: Queue[tuple[StreamingData, int]] = Queue()
        self.timeout = timeout
        self.configuration = configuration
        self.last_counter = -1
        self.max_buffer_size = max_buffer_size
        self.stats = MessageStats()

    def __aiter__(self) -> AsyncIterator[tuple[StreamingData, int]]:
        """Retrieve iterator for collected data

        Returns:

            An iterator over the received streaming data including the number
            of lost messages

        """

        return self

    async def __anext__(self) -> tuple[StreamingData, int]:
        """Retrieve next stream data object in collected data

        Returns:

            A tuple containing:

            - the data of the streaming message and
            - the number of lost streaming messages right before the returned
              streaming message

        """

        if self.queue.qsize() > self.max_buffer_size:
            raise StreamingBufferError(
                f"Maximum buffer size of {self.max_buffer_size} messages "
                "exceeded"
            )

        try:
            return await wait_for(self.queue.get(), self.timeout)
        except TimeoutError as error:
            raise StreamingTimeoutError(
                f"No data received for at least {self.timeout} seconds"
            ) from error

    def on_message_received(self, msg: Message) -> None:
        """Handle received messages

        Args:

            msg:
                The received CAN message

        """

        # Ignore messages with wrong id and “Stop Stream” messages
        if msg.arbitration_id != self.identifier.value or len(msg.data) <= 1:
            return

        data = msg.data
        counter = data[1]
        timestamp = msg.timestamp
        data_bytes = (
            (data[2:4], data[4:6], data[6:8])
            if self.configuration.data_length() == 3
            else (data[2:4], data[4:6])
        )

        values = [
            int.from_bytes(word, byteorder="little") for word in data_bytes
        ]
        assert len(values) == 2 or len(values) == 3

        streaming_data = StreamingData(
            timestamp=timestamp,
            counter=counter,
            values=values,
        )

        # Calculate amount of lost messages
        if self.last_counter < 0:
            self.last_counter = (counter - 1) % 256
        last_counter = self.last_counter
        lost_messages = (counter - last_counter) % 256 - 1
        self.last_counter = counter
        self.stats.lost += lost_messages
        self.stats.retrieved += 1

        self.queue.put_nowait((streaming_data, lost_messages))

    def on_error(self, exc: Exception) -> None:
        """This method is called to handle any exception in the receive thread.

        Args:

            exc:
                The exception causing the thread to stop

        """

        raise NotImplementedError()

    def stop(self) -> None:
        """Stop handling new messages"""

    def reset_stats(self) -> None:
        """Reset the message statistics

        This method resets the amount of lost an retrieved messages used in
        the calculation of the method ``dataloss``. Using this method can be
        useful, if you want to calculate the amount of data loss since a
        specific starting point.

        """

        self.stats.reset()

    def dataloss(self) -> float:
        """Calculate the overall amount of data loss

        Returns:

            The overall amount of data loss as number between 0 (no data loss)
            and 1 (all data lost).

        """

        return self.stats.dataloss()


class StreamingFormat:
    """Support for specifying the `data streaming format`_

    .. _data streaming format:
       https://mytoolit.github.io/Documentation/#block-streaming

    Args:

        *value:
            The value of the streaming format byte

        streaming:
            Specifies if this is a request for a stream of data bytes;
            If this value is not set or set to ``False``, then the request is
            only for a single value (or set of values).

        width:
            Specifies the width of a single value (either 2 or 3 bytes)

        channels:
            Specifies for which channels data should be transmitted or not

        sets:
            Specifies the number of data sets that should be transmitted

            The value 0 stops the stream. Other possible values for the
            number of sets are 1, 3, 6, 10, 15, 20 and 30.

        value_explanations:
            Three strings used to describe the first, second and third data
            value

    """

    data_set = [0, 1, 3, 6, 10, 15, 20, 30]
    """Possible number of data sets"""

    # pylint: disable=too-many-arguments

    def __init__(
        self,
        *value,
        streaming: bool | None = None,
        width: int | None = 2,
        channels: StreamingConfiguration | None = None,
        sets: int | None = None,
        value_explanations: tuple[str, str, str] = (
            "Value 1",
            "Value 2",
            "Value 3",
        ),
    ) -> None:

        def set_part(start, width, number):
            """Store bit pattern number at bit start of the identifier"""

            streaming_ones = 0xFF
            mask = (1 << width) - 1

            # Set all bits for targeted part to 0
            self.value &= (mask << start) ^ streaming_ones
            # Make sure we use the correct number of bits for number
            number = number & mask
            # Set bits to given value
            self.value |= number << start

        self.value_explanations = value_explanations

        if len(value) > 1:
            raise ValueError("More than one positional argument")

        self.value = value[0] if value else 0

        # =============
        # = Streaming =
        # =============

        if streaming:
            set_part(7, 1, int(streaming))

        # =========
        # = Width =
        # =========

        if width is not None:
            if not (isinstance(width, int) and 2 <= width <= 3):
                raise ValueError(f"Unsupported width value: {width}")

            set_part(6, 1, 1 if width == 3 else 0)

        # =================
        # = Active Values =
        # =================

        if channels:
            for shift, part in enumerate(
                [channels.third, channels.second, channels.first]
            ):
                if part is not False:
                    set_part(3 + shift, 1, part)

        # =============
        # = Data Sets =
        # =============

        if sets is not None:
            cls = type(self)

            if sets not in cls.data_set:
                raise ValueError(f"Unsupported number of data sets: {sets}")

            set_part(0, 3, cls.data_set.index(sets))

    # pylint: enable=too-many-arguments

    def __repr__(self) -> str:
        """Retrieve the textual representation of the streaming format

        Returns:

            A string that describes the streaming format

        Examples:

            Get the textual representation of some example streaming formats

            >>> StreamingFormat(width=3,
            ...                 channels=StreamingConfiguration(first=True),
            ...                 sets=15)
            Single Request, 3 Bytes, 15 Data Sets, Read Value 1

            >>> StreamingFormat(0b001, streaming=True)
            Streaming, 2 Bytes, 1 Data Set

            >>> StreamingFormat(0b110111)
            Single Request, 2 Bytes, 30 Data Sets, Read Value 1, Read Value 2

        """

        streaming = self.value >> 7

        data_sets = self.data_sets()
        data_set_explanation = (
            "Stop Stream"
            if data_sets == 0
            else f"{data_sets} Data Set{'' if data_sets == 1 else 's'}"
        )

        parts = [
            "Streaming" if streaming else "Single Request",
            f"{self.data_bytes()} Bytes",
            f"{data_set_explanation}",
        ]

        value_selection = (self.value >> 3) & 0b111

        first = value_selection >> 2
        second = value_selection >> 1 & 1
        third = value_selection & 1

        selected_values = [
            f"Read {value_explanation}"
            for selected, value_explanation in zip(
                (first, second, third), self.value_explanations
            )
            if selected
        ]

        value_explanation = (
            ", ".join(selected_values) if selected_values else ""
        )
        if value_explanation:
            parts.append(value_explanation)

        return ", ".join(parts)

    def data_sets(self) -> int:
        """Get the number of data sets of the streaming format

        Returns:

            The number of data sets

        Examples:

            Retrieve the number of data sets for some example streaming formats

            >>> StreamingFormat(
            ...     width=3,
            ...     channels=StreamingConfiguration(first=True),
            ...     sets=15
            ... ).data_sets()
            15

            >>> StreamingFormat(
            ...     channels=StreamingConfiguration(first=True, second=False),
            ...     sets=3
            ... ).data_sets()
            3

        """

        data_set_bits = self.value & 0b111
        cls = type(self)

        return cls.data_set[data_set_bits]

    def data_bytes(self) -> int:
        """Get the number of data bytes used for a single value

        Returns:

            The number of data bytes that represent a single streaming value

        Examples:

            Retrieve the number of data bytes for some streaming formats

            >>> StreamingFormat(width=3,
            ...                 channels=StreamingConfiguration(first=True),
            ...                 sets=15).data_bytes()
            3

            >>> StreamingFormat(
            ...     channels=StreamingConfiguration(first=True, second=False),
            ...     width=2
            ... ).data_bytes()
            2

        """

        return 3 if (self.value >> 6) & 1 else 2


class StreamingFormatVoltage(StreamingFormat):
    """Support for specifying the streaming format of voltage data

    Args:

        value:
            The value of the streaming format byte

        single:
            Specifies if the request was for a single value or not

        width:
            Specifies the width of a single value (either 2 or 3 bytes)

        channels:
            Specifies for which channels data should be transmitted or not

        sets:
            Specifies the number of data sets that should be transmitted

            The value 0 stops the stream. Other possible values for the
            number of sets are 1, 3, 6, 10, 15, 20 and 30.

    """

    def __init__(self, *arguments, **keyword_arguments) -> None:

        super().__init__(
            *arguments,
            **keyword_arguments,
            value_explanations=("Voltage 1", "Voltage 2", "Voltage 3"),
        )


class StreamingData:
    """Support for storing data of a streaming message

    Args:

        counter:
            The message counter value

        timestamp:
            The message timestamp

        values:
            The streaming values

    Examples:

        Create new streaming data

        >>> StreamingData(values=[1, 2, 3], counter=21, timestamp=1)
        [1, 2, 3]@1 #21

        Streaming data must store either two or three values

        >>> StreamingData(values=[1], counter=21, timestamp=1)
        Traceback (most recent call last):
        ...
        ValueError: Incorrect number of streaming values: 1 (instead of 2 or 3)

        >>> StreamingData(values=[1, 2, 3, 4], counter=21, timestamp=1
        ...              ) # doctest:+ELLIPSIS
        Traceback (most recent call last):
        ...
        ValueError: Incorrect number of ... values: 4 (instead of 2 or 3)

    """

    def __init__(
        self, counter: int, timestamp: float, values: Sequence[float]
    ) -> None:

        if not 2 <= len(values) <= 3:
            raise ValueError(
                f"Incorrect number of streaming values: {len(values)} "
                "(instead of 2 or 3)"
            )

        self.counter = counter
        self.timestamp = timestamp
        self.values = values

    def apply(
        self,
        function: Callable[[float], float],
    ) -> StreamingData:
        """Apply a certain function to the streaming data

        Note:

            This function changes the stored values in the streaming data and
            (as convenience feature) also returns the modified streaming data
            itself. This is useful if you want to use the modified streaming
            as parameter in a function call, i.e. you can use something like
            ``function(stream_data.apply())``.

        Args:

            function:
                The function that should be applied to the streaming data

        Returns:

            The modified streaming data

        Examples:

            Add the constant 10 to some example streaming data

            >>> data = StreamingData(values=[1, 2, 3], counter=21, timestamp=1)
            >>> data.apply(lambda value: value + 10)
            [11, 12, 13]@1 #21
            >>> data.values
            [11, 12, 13]

        """

        updated_values = [function(value) for value in self.values]
        assert len(updated_values) == 2 or len(updated_values) == 3
        self.values = updated_values

        return self

    def __repr__(self):
        """Get the string representation of the stream data

        Examples:

            Get the string representation of some streaming data

            >>> StreamingData(values=[1, 2, 3], counter=21, timestamp=1)
            [1, 2, 3]@1 #21

        """

        return f"{self.values}@{self.timestamp} #{self.counter}"


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



"""Test code for sensory milling head (SMH)"""

# -- Imports ------------------------------------------------------------------

from unittest import main as unittest_main

from icotronic.can.streaming import StreamingConfiguration
from icotronic.cmdline.commander import Commander
from icotronic.config import settings
from icotronic.measurement.sensor import guess_sensor, SensorConfiguration
from icotronic.report.report import Report
from icotronic.test.unit import ExtendedTestRunner
from icotronic.test.production.node import BaseTestCases, TestAttribute

# -- Classes ------------------------------------------------------------------


class TestSMH(BaseTestCases.TestSensorNode):
    """This class contains tests for the milling head sensor (PCB)"""

    @classmethod
    def setUpClass(cls):
        """Set up data for whole test"""

        super().setUpClass()
        cls.report = Report(node="SMH")
        # Guessed sensor types
        for sensor in range(1, settings.smh.channels + 1):
            cls.attributes[f"Sensor {sensor}"] = TestAttribute(str, pdf=True)

    async def _connect(self):
        """Create a connection to the SMH"""

        await super()._connect_node(settings.smh.name)

    async def _disconnect(self):
        """Tear down connection to SMH"""

        await super()._disconnect_node()

    def test__firmware_flash_disconnected(self):
        """Upload bootloader and application into SMH

        Note:

            The additional underscore in the method name that makes sure this
            test case is executed before all other test cases.

        Note:

            The text ``disconnected`` in the method name makes sure that the
            test framework does not initialize a connection.

        """

        self._test_firmware_flash(
            flash_location=settings.smh.firmware.location.flash,
            programmmer_serial_number=(
                settings.smh.programming_board.serial_number
            ),
            chip="BGM121A256V2",
        )

    async def test_eeprom(self):
        """Test if reading and writing the EEPROM works"""

        cls = type(self)

        # ========
        # = Name =
        # ========

        cls.attributes["Name"].value = await self._test_name(settings.smh.name)

        # =========================
        # = Sleep & Advertisement =
        # =========================

        await self._test_eeprom_sleep_advertisement_times()

        # ================
        # = Product Data =
        # ================

        await self._test_eeprom_product_data(settings.smh)

        # ==============
        # = Statistics =
        # ==============

        await self._test_eeprom_statistics(
            settings.smh.production_date,
            settings.smh.batch_number,
        )

    async def test_sensors(self):
        """Test available sensor channels"""

        async def read_streaming_data_amount(length: int):
            async with self.node.open_data_stream(
                StreamingConfiguration(first=True, second=False, third=False)
            ) as stream:
                stream_data = []
                async for data, _ in stream:
                    stream_data.extend(data.values)
                    if len(stream_data) >= length:
                        break

            # Due to the chosen streaming format the code above might have
            # collected one or two additional values. We remove these values
            # here.
            assert len(stream_data) >= length
            additional_values = len(stream_data) - length
            return stream_data[:-additional_values]

        cls = type(self)
        sensors = []

        for test_channel in range(1, settings.smh.channels + 1):
            await self.node.set_sensor_configuration(
                SensorConfiguration(first=test_channel)
            )
            config = await self.node.get_sensor_configuration()
            self.assertEqual(
                config.first,
                test_channel,
                f"Read sensor channel number “{config.first}” does "
                f"not match expected channel number “{test_channel}”",
            )
            stream_data = await read_streaming_data_amount(1000)
            sensor = guess_sensor(stream_data)
            sensors.append(sensor)
            cls.attributes[f"Sensor {test_channel}"].value = str(sensor)

        non_working_sensors = [
            str(sensor_number)
            for sensor_number, sensor in enumerate(sensors, start=1)
            if not sensor.works()
        ]

        if len(non_working_sensors) >= 1:
            if len(non_working_sensors) == 1:
                error_text = f"channel {non_working_sensors.pop()} seems"
            elif len(non_working_sensors) >= 2:
                channels = (
                    ", ".join(non_working_sensors[:-1])
                    + f" & {non_working_sensors[-1]}"
                )
                error_text = f"channels {channels} seem"
            plural = "" if len(non_working_sensors) <= 1 else "s"
            self.assertFalse(
                non_working_sensors,
                f"The sensor{plural} on measurement {error_text} "
                "to not work correctly.",
            )

    def test_power_usage_disconnected(self) -> None:
        """Check power usage in disconnected state"""

        commander = Commander(
            serial_number=settings.smh.programming_board.serial_number,
            chip="BGM121A256V2",
        )

        commander.enable_debug_mode()
        power_usage_mw = commander.read_power_usage()

        expected_maxmimum_usage_mw = 10
        self.assertLess(
            power_usage_mw,
            expected_maxmimum_usage_mw,
            f"Measured power usage of {power_usage_mw} mW is "
            "higher than expected maximum value "
            f"{expected_maxmimum_usage_mw} mW",
        )


# -- Main ---------------------------------------------------------------------


def main():
    """Run production test for Sensory Milling Head (SMH)"""

    unittest_main(
        testRunner=ExtendedTestRunner, module="icotronic.test.production.smh"
    )


if __name__ == "__main__":
    main()



"""Support for MyTooliT protocol CAN messages

For more information, please take a look here:
- https://mytoolit.github.io/Documentation/#mytoolit-communication-protocol
"""

# -- Imports ------------------------------------------------------------------

from __future__ import annotations


from can import Message as CANMessage
from netaddr import EUI

from icotronic.can.calibration import CalibrationMeasurementFormat
from icotronic.can.protocol.command import Command
from icotronic.can.protocol.identifier import Identifier
from icotronic.can.node.id import NodeId
from icotronic.can.status import State
from icotronic.can.streaming import StreamingFormat, StreamingFormatVoltage
from icotronic.utility.data import convert_bytes_to_text

# -- Classes ------------------------------------------------------------------


class Message:
    """Wrapper class for CAN messages

    Usually you will either specify the (PCAN or python-can) message
    directly, or provide identifier and data. If, you decide to specify
    both the message and one of the keyword arguments, then the keyword
    arguments will be used to overwrite specific parts of the message
    argument.

    Additional keyword parameters will be used as arguments for the
    identifier. This way you can specify all the arguments for a message
    without creating the identifier class yourself. If you decide to
    specify both an identifier and keyword arguments for the identifier,
    then the keyword arguments will overwrite the specific parts of the
    provided identifier.

    Args:

        message:

            A python-can message

        identifier:

            A 29 bit CAN identifier

        data:

            An iterable over 8 bit values that stores the payload of the
            message

    Examples:

        Create a message from a python-can message

        >>> from can import Message as CANMessage
        >>> message = Message(
        ...     CANMessage(
        ...         is_extended_id=True,
        ...         arbitration_id=0b0_000000_00000001_00_0_01111_0_00001))
        >>> from re import search
        >>> search('# (.*)', repr(message))[1]
        '[SPU 1 → STH 1, Block: System, Command: Reset, Acknowledge]'

        Create a message using identifier and data

        >>> identifier = Identifier(block=0, block_command=1)
        >>> payload = [0xab, 0xcd]
        >>> message = Message(identifier=identifier, data=payload)

        Create a message using keyword arguments handled by the identifier
        class

        >>> message = Message(block='System', block_command='Reset')

        If you provide both identifier and identifier keyword arguments, then
        the specific keyword arguments overwrite the values of the identifier

        >>> identifier = Identifier(block='System')
        >>> message = Message(identifier=identifier, block='Streaming')
        >>> Identifier(message.id()).block_name()
        'Streaming'


    """

    def __init__(
        self,
        *message: CANMessage,
        identifier: Identifier | None = None,
        data: list[int] | None = None,
        **keyword_arguments: Command | NodeId | None | str | int | bool,
    ) -> None:

        if message:
            can_message = message[0]
            if isinstance(can_message, CANMessage):
                self.message = can_message
            else:
                raise ValueError(
                    "Unsupported object type for argument message: "
                    f"“{type(can_message)}”"
                )
        else:
            self.message = CANMessage(is_extended_id=True)

        if identifier:
            self.message.arbitration_id = (
                identifier if isinstance(identifier, int) else identifier.value
            )

        if keyword_arguments:
            # mypy assumes that all keyword arguments have the same type
            self.message.arbitration_id = Identifier(
                self.id(), **keyword_arguments  # type: ignore
            ).value

        if data:
            self.message.data = bytearray(data)
            self.message.dlc = len(data)

    @property
    def data(self) -> bytearray:
        """Retrieve the message data

        Returns:

            A bytearray containing the data of the message

        Examples:

            Get data from various example messages

            >>> len(Message().data)
            0
            >>> len(Message(data=[5, 6, 7]).data)
            3
            >>> len(Message(data=[1, 2]).data)
            2

        """

        return self.message.data

    @data.setter
    def data(self, data: list[int] | bytearray) -> None:
        """Set the message data to a new value

        Args:

            data:
                The data that should be stored in the message

        Examples:

            Set message data to various values

            >>> message = Message(data=[1])
            >>> message.data[0] = 42
            >>> message.data[0]
            42

            >>> message = Message(data = bytearray([0] * 8))
            >>> message.data[1] = 1
            >>> message.data[2] = 2
            >>> message.data[7] = 7

            >>> message.data[1]
            1
            >>> message.data[3]
            0
            >>> message.data[7]
            7

        """

        self.message.data = bytearray(data)
        self.message.dlc = len(self.data)

    # pylint: disable=too-many-branches, too-many-locals, too-many-statements

    def _data_explanation_system(self) -> str:
        """Retrieve a textual representation of system messages

        Returns:

            A textual description of the data contained in the message

        """

        def mac_address() -> EUI:
            """Convert the message data into a MAC address"""
            return EUI("-".join((f"{byte:0x}" for byte in self.data[7:1:-1])))

        identifier = self.identifier()
        data_explanation = ""

        if identifier.block_command_name() == "Get/Set State":
            if len(self.data) < 1:
                return ""

            data_explanation = repr(State(self.data[0]))

        if identifier.block_command_name() == "Bluetooth":
            subcommand = self.data[0]
            sensor_node_number = self.data[1]
            is_acknowledgment = self.identifier().is_acknowledgment()
            verb = "Return" if is_acknowledgment else "Get"

            if subcommand == 1:
                verb = "Acknowledge" if is_acknowledgment else "Request"
                data_explanation = f"{verb} Bluetooth activation"
            elif subcommand == 2:
                data_explanation = f"{verb} number of available nodes"
                if is_acknowledgment:
                    number_nodes_text = convert_bytes_to_text(self.data[2:])
                    data_explanation += ": "
                    try:
                        number_nodes = int(number_nodes_text)
                        data_explanation += f"{number_nodes}"
                    except ValueError:
                        data_explanation += (
                            "Unable to convert text "
                            f"“{number_nodes_text}” to number"
                        )

            elif subcommand in (5, 6):
                part = "first" if subcommand == 5 else "second"
                data_explanation = (
                    f"{verb} {part} part of name of node "
                    f"with node number “{sensor_node_number}”"
                )
                if is_acknowledgment and len(self.data) >= 2:
                    name = convert_bytes_to_text(self.data[2:])
                    data_explanation += f": “{name}”"
            elif subcommand == 7:
                info = (
                    "Acknowledge connection request"
                    if is_acknowledgment
                    else "Request connection"
                )
                data_explanation = (
                    f"{info} to node with node number “{sensor_node_number}”"
                )
            elif subcommand == 8:
                data_explanation = f"{verb} Bluetooth connection status"
                if is_acknowledgment and len(self.data) >= 3:
                    connected = bool(self.data[2])
                    data_explanation += (
                        f": {'Connected' if connected else 'Not connected'}"
                    )
            elif subcommand == 9:
                verb = "Acknowledge" if is_acknowledgment else "Request"
                data_explanation = f"{verb} Bluetooth deactivation"
            elif subcommand == 10:
                verb = "Return" if is_acknowledgment else "Get"
                data_explanation = f"{verb} Bluetooth send counter"
                if is_acknowledgment and len(self.data) >= 8:
                    counter = int.from_bytes(self.data[2:], byteorder="big")
                    data_explanation += f": {counter}"
            elif subcommand == 12:
                data_explanation = (
                    f"{verb} RSSI of node "
                    f"with node number “{sensor_node_number}”"
                )
                if is_acknowledgment and len(self.data) >= 3:
                    rssi = int.from_bytes(
                        self.data[2:3], byteorder="little", signed=True
                    )
                    data_explanation += f": {rssi}"
            elif subcommand == 14:
                data_explanation = "Write energy mode reduced"
                if len(self.data) >= 8:
                    time_normal_to_reduced_ms = int.from_bytes(
                        self.data[2:6], byteorder="little"
                    )
                    advertisement_time_eeprom_to_ms = 0.625
                    advertisement_time = (
                        int.from_bytes(self.data[6:], byteorder="little")
                        * advertisement_time_eeprom_to_ms
                    )
                    data_explanation += ": " + ", ".join([
                        f"⟳ {time_normal_to_reduced_ms} ms",
                        f"📢 {advertisement_time} ms",
                    ])
            elif subcommand == 17:
                data_explanation = (
                    f"{verb} MAC address of node "
                    f"with node number “{sensor_node_number}”"
                )
                if is_acknowledgment and len(self.data) >= 8:
                    data_explanation += f": {mac_address()}"
            elif subcommand == 18:
                verb = (
                    "Acknowledge connection request"
                    if is_acknowledgment
                    else "Request connection"
                )
                data_explanation = (
                    f"{verb} to node with MAC address “{mac_address()}”"
                )

        return data_explanation

    # pylint: enable=too-many-branches, too-many-locals, too-many-statements

    def _data_explanation_streaming(self) -> str:
        """Retrieve a textual representation of streaming messages

        Returns:

            A textual description of the data contained in the message

        """

        if len(self.data) < 1:
            return ""

        identifier = self.identifier()
        data_explanation = ""

        block_command = identifier.block_command_name()

        StreamingFormatClass = (
            StreamingFormat
            if block_command == "Data"
            else (
                StreamingFormatVoltage
                if block_command == "Voltage"
                else StreamingFormat
            )
        )
        streaming_format: StreamingFormat = StreamingFormatClass(self.data[0])
        data_explanation += repr(streaming_format)

        if identifier.is_acknowledgment() and len(self.data) >= 2:
            sequence_counter = self.data[1]
            explanations = [f"Sequence Counter: {sequence_counter}"]
            size_value = streaming_format.data_bytes()
            values = [
                int.from_bytes(
                    self.data[start : start + size_value], byteorder="little"
                )
                for start in range(2, len(self.data), size_value)
            ]
            explanations.extend([
                f"{explanation}: {value}"
                for explanation, value in zip(
                    streaming_format.value_explanations, values
                )
            ])
            data_explanation += ", " + ", ".join(explanations)

        return data_explanation

    def _data_explanation_configuration(self) -> str:
        """Retrieve a textual representation of configuration messages

        Returns:

            A textual description of the data contained in the message

        """

        identifier = self.identifier()
        data_explanation = ""

        if (
            identifier.block_command_name() == "Calibration Measurement"
            and len(self.data) >= 4
        ):
            return repr(CalibrationMeasurementFormat(self.data))

        return data_explanation

    def _data_explanation_eeprom(self) -> str:
        """Retrieve a textual representation of EEPROM messages

        Returns:

            A textual description of the data contained in the message

        """

        identifier = self.identifier()
        data_explanation = ""

        if identifier.block_command_name() == "Read Write Request Counter":
            counter = int.from_bytes(self.data[4:], "little")
            data_explanation = f"EEPROM Write Requests: {counter}"
        elif (
            identifier.block_command_name() in {"Read", "Write"}
            and len(self.data) >= 8
        ):
            page = self.data[0]
            offset = self.data[1]
            length = self.data[2]
            data = list(self.data[4 : 4 + min(length, 4)])
            is_acknowledgment = self.identifier().is_acknowledgment()

            info = "Acknowledge request" if is_acknowledgment else "Request"
            verb = identifier.block_command_name().lower()
            data_explanation = (
                f"{info} to {verb} {length} bytes at page {page} "
                f"with offset {offset}"
            )
            read_request = verb == "read" and not is_acknowledgment
            if not read_request:
                data_explanation += f": {data}"

        return data_explanation

    def _data_explanation(self) -> str:
        """Retrieve a textual representation of the message data

        Returns:

            A textual description of the data contained in the message

        """

        if len(self.data) <= 0:
            return ""

        identifier = self.identifier()

        if identifier.block_name() == "System":
            return self._data_explanation_system()

        if identifier.block_name() == "Streaming":
            return self._data_explanation_streaming()

        if identifier.block_name() == "Configuration":
            return self._data_explanation_configuration()

        if identifier.block_name() == "EEPROM":
            return self._data_explanation_eeprom()

        return ""

    def __repr__(self) -> str:
        """Get a textual representation of the current message

        Returns:

            A text that shows the various attributes of the current message

        Examples:

            Retrieve the textual representation of various example messages

            >>> identifier = Identifier(block=0, block_command=1,
            ...                         request=True, error=False,
            ...                         sender=1, receiver=14)
            >>> message = Message(identifier=identifier, data=[0xfe, 0xfe])
            >>> message # doctest:+NORMALIZE_WHITESPACE
            0b00000000000000110000001001110 2 0xfe 0xfe
            # [STH 1 → STH 14, Block: System, Command: Reset, Request]

            >>> from re import search
            >>> representation = repr(
            ...     Message(block='Streaming',
            ...             block_command='Data',
            ...             sender='SPU 1',
            ...             receiver='STH 1',
            ...             request=True))
            >>> search('# (.*)', representation)[1]
            '[SPU 1 → STH 1, Block: Streaming, Command: Data, Request]'

            >>> representation = repr(
            ...     Message(block='System',
            ...             block_command='Bluetooth',
            ...             sender='SPU 1',
            ...             receiver='STU 1',
            ...             request=True,
            ...             data=[1] + [0]*7))
            >>> match = search('# (.*)', representation)[1]
            >>> match # doctest:+NORMALIZE_WHITESPACE
            '[SPU 1 → STU 1, Block: System, Command: Bluetooth, Request]
             (Request Bluetooth activation)'

            >>> representation = repr(
            ...     Message(block='System',
            ...             block_command='Get/Set State',
            ...             sender='SPU 2',
            ...             receiver='STU 1',
            ...             request=True,
            ...             data=[0b10_0_101]))
            >>> match = search('# (.*)', representation)[1]
            >>> match # doctest:+NORMALIZE_WHITESPACE
            '[SPU 2 → STU 1, Block: System, Command: Get/Set State, Request]
             (Get State, Location: Application, State: Operating)'

        """

        identifier = self.identifier()
        data_explanation = self._data_explanation()

        explanation = (
            f"{identifier} ({data_explanation})"
            if data_explanation
            else repr(identifier)
        )

        data_representation = " ".join(
            [hex(self.data[byte]) for byte in range(len(self.data))]
        )
        bit_values = [
            f"0b{identifier.value:029b}",
            str(len(self.data)),
            data_representation,
        ]
        # Filter empty string, since otherwise there might be an additional
        # space at the end of the representation for empty data
        bit_representation = " ".join(filter(None, bit_values))

        return f"{bit_representation} # {explanation}"

    def acknowledge(self, error: bool = False) -> Message:
        """Returns an acknowledgment message object for this message

        In the acknowledgment message receiver and sender will be swapped and
        the request (acknowledge) bit will be set to 0 (acknowledge). The data
        of the acknowledgment message will be the same as in the original
        message.

        Args:

            error:
                Specifies if the acknowledgment message is an error
                acknowledgment or not

        Returns:

            An acknowledgment message for the current message

        Examples:

            >>> identifier = Identifier(block=0, block_command=1, sender=5,
            ...                         receiver=10)
            >>> message = Message(identifier=identifier, data=[0xaa])
            >>> message.acknowledge() # doctest:+NORMALIZE_WHITESPACE
            0b00000000000000100001010000101 1 0xaa
            # [STH 10 → STH 5, Block: System, Command: Reset, Acknowledge]

        """

        return Message(
            identifier=self.identifier().acknowledge(),
            error=error,
            data=list(self.data),
        )

    def id(self) -> int:
        """Retrieve the ID of the message

        Returns:

            The 29 bit CAN identifier of the message

        Examples:

            Check the identifier of an example message

            >>> Message(block='Configuration').id(
            ...        ) == 0b0_1010_000000000000000000000000
            True

        """

        return self.message.arbitration_id

    def identifier(self) -> Identifier:
        """Retrieve an identifier object for the ID of the message

        Returns:

            An identifier object representing the ID of the message

        Examples:

            Get the textual representation of an example message

            >>> Message(block='System', block_command='Reset', request=True,
            ...         sender='SPU 1', receiver='STH 2').identifier()
            [SPU 1 → STH 2, Block: System, Command: Reset, Request]

        """

        return Identifier(self.id())

    def to_python_can(self) -> CANMessage:
        """Retrieve a python-can message object for this message

        Returns:

            A message object of the python-can API

        Examples:

            Check the python-can message representation of an example message

            >>> message = Message(block='EEPROM', sender='SPU 1', data=[1,2,3])
            >>> can_message = message.to_python_can()
            >>> can_message # doctest:+NORMALIZE_WHITESPACE
            can.Message(timestamp=0.0, arbitration_id=0xf4003c0,
                        is_extended_id=True, dlc=3, data=[0x1, 0x2, 0x3])

            >>> message.id() == can_message.arbitration_id
            True

        """

        return self.message


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



"""Support for measuring data"""

# -- Exports ------------------------------------------------------------------

from .acceleration import convert_raw_to_g, ratio_noise_max
from .storage import Storage, StorageData



"""Support for reading voltage data"""

# -- Imports ------------------------------------------------------------------

from icotronic.measurement.constants import ADC_MAX_VALUE

# -- Functions ----------------------------------------------------------------


def convert_raw_to_supply_voltage(
    voltage_raw: int, reference_voltage: float = 3.3
) -> float:
    """Convert a raw 2 byte ADC value to a voltage value

    Args:

        voltage_raw:
            A 2 byte supply voltage value (returned by the `Streaming` command)

        reference_voltage:
            The ADC reference voltage

    Returns:

        The supply voltage in volts

    Examples:

        Convert some ADC values into supply voltage values

        >>> voltage = convert_raw_to_supply_voltage(11000)
        >>> 3.15 < voltage < 3.16
        True

        >>> voltage = convert_raw_to_supply_voltage(2**15,
        ...                                         reference_voltage=1.8)
        >>> 5.12 < voltage < 5.14
        True

    """

    if voltage_raw <= 0:
        return 0

    # The value below is the result of the voltage divider circuit, which
    # uses a 4.7 kΩ and 1 kΩ resistor: (470 kΩ + 100 kΩ) / 100 kΩ = 5.7
    # If you want to know why we used these resistor values, then please
    # take a look
    # [here](https://en.wikipedia.org/wiki/E_series_of_preferred_numbers).
    voltage_divider_factor = 5.7
    return (
        voltage_raw
        * voltage_divider_factor
        * reference_voltage
        / ADC_MAX_VALUE
    )


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



"""Support for acceleration based sensor nodes (SHA and STH)"""

# -- Imports ------------------------------------------------------------------

from functools import partial
from collections.abc import Callable

from icotronic.can.calibration import CalibrationMeasurementFormat
from icotronic.can.node.eeprom.sth import STHEEPROM
from icotronic.can.protocol.message import Message
from icotronic.can.node.sensor import SensorNode
from icotronic.can.node.spu import SPU
from icotronic.measurement.acceleration import convert_raw_to_g
from icotronic.measurement.constants import ADC_MAX_VALUE

# -- Classes ------------------------------------------------------------------


class STH(SensorNode):
    """Communicate and control a connected SHA or STH

    Args:

        spu:
            The SPU object used to communicate with the node

    """

    def __init__(self, spu: SPU) -> None:

        super().__init__(spu, STHEEPROM)

    # ---------------------------
    # - Calibration Measurement -
    # ---------------------------

    async def _acceleration_self_test(
        self, activate: bool = True, dimension: str = "x"
    ) -> None:
        """Activate/Deactivate the accelerometer self test

        Args:

            activate:
                Either ``True`` to activate the self test or ``False`` to
                deactivate the self test

            dimension:
                The dimension (x=1, y=2, z=3) for which the self test should be
                activated/deactivated.

        """

        node = self.id
        method = "Activate" if activate else "Deactivate"

        try:
            dimension_number = "xyz".index(dimension) + 1
        except ValueError as error:
            raise ValueError(
                f"Invalid dimension value: “{dimension}”"
            ) from error

        message = Message(
            block="Configuration",
            block_command="Calibration Measurement",
            sender=self.spu.id,
            receiver=node,
            request=True,
            data=CalibrationMeasurementFormat(
                set=True,
                element="Data",
                method=method,
                dimension=dimension_number,
            ).data,
        )

        await self.spu.request(
            message,
            description=(
                f"{method.lower()} self test of {dimension}-axis of “{node}”"
            ),
        )

    async def activate_acceleration_self_test(
        self, dimension: str = "x"
    ) -> None:
        """Activate self test of STH accelerometer

        Args:

            dimension:
                The dimension (``x``, ``y`` or ``z``) for which the self test
                should be activated.

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Activate and deactivate acceleration self-test

            >>> async def test_self_test():
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is
            ...         # available
            ...         async with stu.connect_sensor_node(0, STH) as sth:
            ...             await sth.activate_acceleration_self_test()
            ...             await sth.deactivate_acceleration_self_test()
            >>> run(test_self_test())

        """

        await self._acceleration_self_test(activate=True, dimension=dimension)

    async def deactivate_acceleration_self_test(
        self, dimension: str = "x"
    ) -> None:
        """Deactivate self test of STH accelerometer

        Args:

            dimension:
                The dimension (``x``, ``y`` or ``z``) for which the self test
                should be deactivated.

        """

        await self._acceleration_self_test(activate=False, dimension=dimension)

    async def get_acceleration_voltage(
        self, dimension: str = "x", reference_voltage: float = 3.3
    ) -> float:
        """Retrieve the current acceleration voltage in Volt

        Args:

            dimension:
                The dimension (x=1, y=2, z=3) for which the acceleration
                voltage should be measured

            reference_voltage:
                The reference voltage for the ADC in Volt

        Returns:

            The voltage of the acceleration sensor in Volt

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Read the acceleration voltage of STH 1

            >>> async def read_acceleration_voltage():
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         async with stu.connect_sensor_node(0, STH) as sth:
            ...             before = await sth.get_acceleration_voltage()
            ...             await sth.activate_acceleration_self_test()
            ...             between = await sth.get_acceleration_voltage()
            ...             await sth.deactivate_acceleration_self_test()
            ...             after = await sth.get_acceleration_voltage()
            ...         return (before, between, after)
            >>> before, between, after = run(read_acceleration_voltage())
            >>> before < between and after < between
            True

        """

        try:
            dimension_number = "xyz".index(dimension) + 1
        except ValueError as error:
            raise ValueError(
                f"Invalid dimension value: “{dimension}”"
            ) from error

        node = self.id
        message = Message(
            block="Configuration",
            block_command="Calibration Measurement",
            sender=self.spu.id,
            receiver=node,
            request=True,
            data=CalibrationMeasurementFormat(
                set=True,
                element="Data",
                method="Measure",
                reference_voltage=reference_voltage,
                dimension=dimension_number,
            ).data,
        )

        response = await self.spu.request(
            message, description=f"retrieve acceleration voltage of “{node}”"
        )

        adc_value = int.from_bytes(response.data[4:], "little")
        return adc_value / ADC_MAX_VALUE * reference_voltage

    async def get_acceleration_sensor_range_in_g(self) -> int:
        """Retrieve the maximum acceleration sensor range in multiples of g₀

        - For a ±100 g₀ sensor this method returns 200 (100 + abs(-100)).
        - For a ±50 g₀ sensor this method returns 100 (50 + abs(-50)).

        For this to work correctly the EEPROM value of the x-axis acceleration
        offset in the EEPROM has to be set.

        Returns:

            Range of current acceleration sensor in multiples of earth’s
            gravitation

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection
            >>> from icotronic.can.node.sth import STH

            Write and read the acceleration offset of STH 1

            >>> async def read_sensor_range():
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         async with stu.connect_sensor_node(0, STH) as sth:
            ...             return (await
            ...                     sth.get_acceleration_sensor_range_in_g())
            >>> sensor_range = run(read_sensor_range())
            >>> 0 < sensor_range <= 200
            True

        """

        assert isinstance(self.eeprom, STHEEPROM)
        return round(
            abs(await self.eeprom.read_x_axis_acceleration_offset()) * 2
        )

    async def get_acceleration_conversion_function(
        self,
    ) -> Callable[[float], float]:
        """Retrieve function to convert raw sensor data into g

        Returns:

            A function that converts 16 bit raw values from the STH into
            multiples of earth’s gravitation (g)

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection
            >>> from icotronic.can.node.sth import STH

            Convert a raw ADC value into multiples of g

            >>> async def read_sensor_values():
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         async with stu.connect_sensor_node(0, STH) as sth:
            ...             convert_to_g = (await
            ...                 sth.get_acceleration_conversion_function())
            ...             data = await sth.get_streaming_data_single()
            ...             before = list(data.values)
            ...             data.apply(convert_to_g)
            ...             return before, data.values
            >>> before, after = run(read_sensor_values())
            >>> all([0 <= value <= 2**16 for value in before])
            True
            >>> all([-100 <= value <= 100 for value in after])
            True

        """

        sensor_range = await self.get_acceleration_sensor_range_in_g()
        return partial(convert_raw_to_g, max_value=sensor_range)


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import run_docstring_examples

    run_docstring_examples(
        STH.get_acceleration_conversion_function,
        globals(),
        verbose=True,
    )



"""Configuration support for ICOtronic tools (e.g. ICOn, production tests)"""

# -- Exports ------------------------------------------------------------------

from .config import ConfigurationUtility, settings



"""Support code for changing the style of PDF (test reports)"""

# -- Imports ------------------------------------------------------------------

from reportlab.rl_config import (  # pylint: disable=no-name-in-module
    canvas_basefontname,
)
from reportlab.lib.fonts import tt2ps
from reportlab.lib.styles import ListStyle, ParagraphStyle, StyleSheet1
from reportlab.lib.colors import black
from reportlab.lib.enums import TA_CENTER

# -- Attributes ---------------------------------------------------------------

base_font = canvas_basefontname
base_font_bold = tt2ps(canvas_basefontname, 1, 0)
base_font_italic = tt2ps(canvas_basefontname, 0, 1)
base_font_bold_italic = tt2ps(canvas_basefontname, 1, 1)

# -- Functions ----------------------------------------------------------------


def get_style_sheet() -> StyleSheet1:
    """Get the default style sheet for PDF reports

    This is more or less a slightly modified version of the
    function ``getSampleStyleSheet`` shipped with ReportLab.

    Returns:

        Style sheet for PDF report

    """

    stylesheet = StyleSheet1()

    stylesheet.add(
        ParagraphStyle(
            name="Normal", fontName=base_font, fontSize=10, leading=12
        )
    )

    stylesheet.add(
        ParagraphStyle(
            name="BodyText", parent=stylesheet["Normal"], spaceBefore=6
        )
    )
    stylesheet.add(
        ParagraphStyle(
            name="Italic",
            parent=stylesheet["BodyText"],
            fontName=base_font_italic,
        )
    )

    stylesheet.add(
        ParagraphStyle(
            name="Heading1",
            parent=stylesheet["Normal"],
            fontName=base_font_bold,
            fontSize=18,
            leading=22,
            spaceAfter=6,
        ),
        alias="h1",
    )

    stylesheet.add(
        ParagraphStyle(
            name="Title",
            parent=stylesheet["Normal"],
            fontName=base_font_bold,
            fontSize=18,
            leading=22,
            alignment=TA_CENTER,
            spaceAfter=6,
        ),
        alias="title",
    )

    stylesheet.add(
        ParagraphStyle(
            name="Heading2",
            parent=stylesheet["Normal"],
            fontName=base_font_bold,
            fontSize=14,
            leading=18,
            spaceBefore=12,
            spaceAfter=6,
        ),
        alias="h2",
    )

    stylesheet.add(
        ParagraphStyle(
            name="Heading3",
            parent=stylesheet["Normal"],
            fontName=base_font_bold,
            fontSize=12,
            leading=14,
            spaceBefore=12,
            spaceAfter=6,
        ),
        alias="h3",
    )

    stylesheet.add(
        ParagraphStyle(
            name="Heading4",
            parent=stylesheet["Normal"],
            fontName=base_font_bold_italic,
            fontSize=10,
            leading=12,
            spaceBefore=10,
            spaceAfter=4,
        ),
        alias="h4",
    )

    stylesheet.add(
        ParagraphStyle(
            name="Heading5",
            parent=stylesheet["Normal"],
            fontName=base_font_bold,
            fontSize=9,
            leading=10.8,
            spaceBefore=8,
            spaceAfter=4,
        ),
        alias="h5",
    )

    stylesheet.add(
        ParagraphStyle(
            name="Heading6",
            parent=stylesheet["Normal"],
            fontName=base_font_bold,
            fontSize=7,
            leading=8.4,
            spaceBefore=6,
            spaceAfter=2,
        ),
        alias="h6",
    )

    stylesheet.add(
        ParagraphStyle(
            name="Bullet",
            parent=stylesheet["Normal"],
            firstLineIndent=0,
            spaceBefore=3,
        ),
        alias="bu",
    )

    stylesheet.add(
        ParagraphStyle(
            name="Definition",
            parent=stylesheet["Normal"],
            firstLineIndent=0,
            leftIndent=36,
            bulletIndent=0,
            spaceBefore=6,
            bulletFontName=base_font_bold_italic,
        ),
        alias="df",
    )

    stylesheet.add(
        ParagraphStyle(
            name="Code",
            parent=stylesheet["Normal"],
            fontName="Courier",
            fontSize=8,
            leading=8.8,
            firstLineIndent=0,
            leftIndent=36,
            hyphenationLang="",
        )
    )

    stylesheet.add(
        ListStyle(
            name="UnorderedList",
            parent=None,
            leftIndent=18,
            rightIndent=0,
            bulletAlign="left",
            bulletType="1",
            bulletColor=black,
            bulletFontName="Helvetica",
            bulletFontSize=12,
            bulletOffsetY=0,
            bulletDedent="auto",
            bulletDir="ltr",
            bulletFormat=None,
            start=None,
        ),
        alias="ul",
    )

    stylesheet.add(
        ListStyle(
            name="OrderedList",
            parent=None,
            leftIndent=18,
            rightIndent=0,
            bulletAlign="left",
            bulletType="1",
            bulletColor=black,
            bulletFontName="Helvetica",
            bulletFontSize=12,
            bulletOffsetY=0,
            bulletDedent="auto",
            bulletDir="ltr",
            bulletFormat=None,
            start=None,
        ),
        alias="ol",
    )

    return stylesheet



"""Support code to add PDF images to PDF (test reports)

Original Source: https://stackoverflow.com/questions/31712386
"""

# -- Imports ------------------------------------------------------------------

from pdfrw import PdfReader, PdfDict
from pdfrw.buildxobj import pagexobj
from pdfrw.toreportlab import makerl
from reportlab.platypus import Flowable

# -- Class --------------------------------------------------------------------


class PDFImage(Flowable):
    """Create a flowable object from a PDF

    Args:

        filepath:
            A path to the PDF object this object represents

        width:
            The width of the PDF object

        height:
            The height of the PDF object

    """

    def __init__(self, filepath, width=200, height=200):

        super().__init__()

        self.width = width
        self.height = height

        with open(filepath, "rb") as pdf_data:
            (page,) = PdfReader(pdf_data).pages
            self.data = pagexobj(page)

    # pylint: disable=unused-argument

    def wrap(self, availWidth, availHeight):
        """Wrap the PDF object according to the given dimensions

        Args:

            availWidth:
                The width of the wrapped PDF object

            availHeight:
                The height of the wrapped PDF object

        Returns:

            The dimensions of the wrapped object

        """
        return self.width, self.height

    # pylint: enable=unused-argument

    def drawOn(self, canvas, x, y, _=0):
        """Draw the PDF object on the given canvas

        Args:

            canvas:
                The canvas where the flowable should be placed

            x:
                The x position of the PDF inside the canvas

            y:
                The y position of the PDF inside the canvas

            _:
                Ignored parameter that stores offset data for the alignment

        """

        canvas.saveState()

        data = self.data
        if isinstance(data, PdfDict):
            x_scale = self.width / data.BBox[2]
            y_scale = self.height / data.BBox[3]
            canvas.translate(x, y)
            canvas.scale(x_scale, y_scale)
            canvas.doForm(makerl(canvas, data))
        else:
            canvas.drawImage(
                self.img_data,  # pylint: disable=no-member
                x,
                y,
                self.width,
                self.height,
            )

        canvas.restoreState()



"""Read and write EEPROM data of ICOtronic sensory tool holders"""

# -- Imports ------------------------------------------------------------------

from icotronic.can.node.eeprom.sensor import SensorNodeEEPROM

# -- Sensor -------------------------------------------------------------------


class STHEEPROM(SensorNodeEEPROM):
    """Read and write EEPROM data of sensory tool holders"""

    # ===============
    # = Calibration =
    # ===============

    async def read_x_axis_acceleration_slope(self) -> float:
        """Retrieve the acceleration slope of the x-axis from the EEPROM

        Returns:

            The x-axis acceleration slope of the STH

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection
            >>> from icotronic.can.node.sth import STH

            Read x-axis acceleration slope of an STH with node number 0

            >>> async def read_x_axis_acceleration_slope():
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         async with stu.connect_sensor_node(0, STH) as sth:
            ...             return ( await
            ...                 sth.eeprom.read_x_axis_acceleration_slope())
            >>> x_axis_acceleration_slope = run(
            ...     read_x_axis_acceleration_slope())
            >>> isinstance(x_axis_acceleration_slope, float)
            True

        """

        return await self.read_float(address=8, offset=0)

    async def write_x_axis_acceleration_slope(self, slope: float) -> None:
        """Write the acceleration slope of the x-axis to the EEPROM

        Args:

            slope:
                The addition to the acceleration value for one step of the ADC
                in multiples of g₀

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from math import isclose
            >>> from icotronic.can.connection import Connection
            >>> from icotronic.can.node.sth import STH

            Write and read the acceleration slope of an STH with node number 0

            >>> async def write_read_x_axis_acceleration_slope(slope):
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         async with stu.connect_sensor_node(0, STH) as sth:
            ...             await (
            ...                 sth.eeprom.write_x_axis_acceleration_slope(
            ...                     slope))
            ...             return await (
            ...                 sth.eeprom.read_x_axis_acceleration_slope())
            >>> adc_max = 0xffff
            >>> acceleration_difference_max = 200
            >>> slope = acceleration_difference_max / adc_max
            >>> slope_read = run(write_read_x_axis_acceleration_slope(slope))
            >>> isclose(slope, slope_read)
            True

        """

        await self.write_float(address=8, offset=0, value=slope)

    async def read_x_axis_acceleration_offset(self) -> float:
        """Retrieve the acceleration offset of the x-axis from the EEPROM

        Returns:

            The acceleration offset of the x-axis of STH 1

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from platform import system
            >>> from icotronic.can.connection import Connection
            >>> from icotronic.can.node.sth import STH

            Read the acceleration offset of an STH with sensor node number 0

            >>> async def read_x_axis_acceleration_offset():
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         async with stu.connect_sensor_node(0, STH) as sth:
            ...             return (await
            ...                 sth.eeprom.read_x_axis_acceleration_offset())
            >>> x_axis_acceleration_offset = run(
            ...     read_x_axis_acceleration_offset())
            >>> isinstance(x_axis_acceleration_offset, float)
            True

        """

        return await self.read_float(address=8, offset=4)

    async def write_x_axis_acceleration_offset(self, offset: int) -> None:
        """Write the acceleration offset of the x-axis to the EEPROM

        Args:

            offset:
                The (negative) offset of the acceleration value in multiples
                of g₀

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from math import isclose
            >>> from icotronic.can.connection import Connection
            >>> from icotronic.can.node.sth import STH

            Write and read the acceleration offset of an STH with number 0

            >>> async def write_read_x_axis_acceleration_offset(offset):
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         async with stu.connect_sensor_node(0, STH) as sth:
            ...             await sth.eeprom.write_x_axis_acceleration_offset(
            ...                 offset)
            ...             return (await
            ...                 sth.eeprom.read_x_axis_acceleration_offset())
            >>> acceleration_difference_max = 200
            >>> offset = -(acceleration_difference_max/2)
            >>> offset_read = run(
            ...     write_read_x_axis_acceleration_offset(offset))
            >>> isclose(offset, offset_read)
            True

        """

        await self.write_float(address=8, offset=4, value=offset)

    async def read_y_axis_acceleration_slope(self) -> float:
        """Retrieve the acceleration slope of the y-axis from the EEPROM

        Returns:

            The y-axis acceleration slope of STH 1

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection
            >>> from icotronic.can.node.sth import STH

            Read the acceleration slope in the y direction of an STH

            >>> async def read_y_axis_acceleration_slope():
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         async with stu.connect_sensor_node(0, STH) as sth:
            ...             return (await
            ...                 sth.eeprom.read_y_axis_acceleration_slope())
            >>> y_axis_acceleration_slope = run(
            ...     read_y_axis_acceleration_slope())
            >>> isinstance(y_axis_acceleration_slope, float)
            True

        """

        return await self.read_float(address=8, offset=8)

    async def write_y_axis_acceleration_slope(self, slope: float) -> None:
        """Write the acceleration slope of the y-axis to the EEPROM

        Args:

            slope:
                The addition to the acceleration value for one step of the ADC
                in multiples of g₀

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from math import isclose
            >>> from icotronic.can.connection import Connection
            >>> from icotronic.can.node.sth import STH

            Write and read the acceleration slope of STH 1

            >>> async def write_read_y_axis_acceleration_slope(slope):
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         async with stu.connect_sensor_node(0, STH) as sth:
            ...             await (
            ...                 sth.eeprom.write_y_axis_acceleration_slope(
            ...                     slope))
            ...             return (await
            ...                 sth.eeprom.read_y_axis_acceleration_slope())
            >>> adc_max = 0xffff
            >>> acceleration_difference_max = 200
            >>> slope = acceleration_difference_max / adc_max
            >>> slope_read = run(write_read_y_axis_acceleration_slope(slope))
            >>> isclose(slope, slope_read)
            True

        """

        await self.write_float(address=8, offset=8, value=slope)

    async def read_y_axis_acceleration_offset(self) -> float:
        """Retrieve the acceleration offset of the y-axis from the EEPROM

        Returns:

            The acceleration offset of the y-axis of STH 1

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection
            >>> from icotronic.can.node.sth import STH

            Read the acceleration offset of STH 1

            >>> async def read_y_axis_acceleration_offset():
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         async with stu.connect_sensor_node(0, STH) as sth:
            ...             return (await
            ...                 sth.eeprom.read_y_axis_acceleration_offset())
            >>> y_axis_acceleration_offset = run(
            ...     read_y_axis_acceleration_offset())
            >>> isinstance(y_axis_acceleration_offset, float)
            True

        """

        return await self.read_float(address=8, offset=12)

    async def write_y_axis_acceleration_offset(self, offset: int) -> None:
        """Write the acceleration offset of the y-axis to the EEPROM

        Args:

            offset:
                The (negative) offset of the acceleration value in multiples
                of g₀

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from math import isclose
            >>> from icotronic.can.connection import Connection
            >>> from icotronic.can.node.sth import STH

            Write and read the acceleration offset of STH 1

            >>> async def write_read_y_axis_acceleration_offset(offset):
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         async with stu.connect_sensor_node(0, STH) as sth:
            ...             await sth.eeprom.write_y_axis_acceleration_offset(
            ...                 offset)
            ...             return (await
            ...                   sth.eeprom.read_y_axis_acceleration_offset())
            >>> acceleration_difference_max = 200
            >>> offset = -(acceleration_difference_max/2)
            >>> offset_read = run(
            ...     write_read_y_axis_acceleration_offset(offset))
            >>> isclose(offset, offset_read)
            True

        """

        await self.write_float(address=8, offset=12, value=offset)

    async def read_z_axis_acceleration_slope(self) -> float:
        """Retrieve the acceleration slope of the z-axis from the EEPROM

        Returns:

            The z-axis acceleration slope of the STH

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection
            >>> from icotronic.can.node.sth import STH

            Read the acceleration slope in the z direction of STH 1

            >>> async def read_z_axis_acceleration_slope():
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         async with stu.connect_sensor_node(0, STH) as sth:
            ...             return (await
            ...                    sth.eeprom.read_z_axis_acceleration_slope())
            >>> z_axis_acceleration_slope = run(
            ...     read_z_axis_acceleration_slope())
            >>> isinstance(z_axis_acceleration_slope, float)
            True

        """

        return await self.read_float(address=8, offset=16)

    async def write_z_axis_acceleration_slope(self, slope: float) -> None:
        """Write the acceleration slope of the z-axis to the EEPROM

        Args:

            slope:
                The addition to the acceleration value for one step of the ADC
                in multiples of g₀

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from math import isclose
            >>> from icotronic.can.connection import Connection
            >>> from icotronic.can.node.sth import STH

            Write and read the acceleration slope of STH 1

            >>> async def write_read_z_axis_acceleration_slope(slope):
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         async with stu.connect_sensor_node(0, STH) as sth:
            ...             (await
            ...                 sth.eeprom.write_z_axis_acceleration_slope(
            ...                 slope))
            ...             return (await
            ...                 sth.eeprom.read_z_axis_acceleration_slope())
            >>> adc_max = 0xffff
            >>> acceleration_difference_max = 200
            >>> slope = acceleration_difference_max / adc_max
            >>> slope_read = run(write_read_z_axis_acceleration_slope(slope))
            >>> isclose(slope, slope_read)
            True

        """

        await self.write_float(address=8, offset=16, value=slope)

    async def read_z_axis_acceleration_offset(self) -> float:
        """Retrieve the acceleration offset of the z-axis from the EEPROM

        Returns:

            The acceleration offset of the z-axis of STH 1

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection
            >>> from icotronic.can.node.sth import STH

            Read the acceleration offset of an STH

            >>> async def read_z_axis_acceleration_offset():
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         async with stu.connect_sensor_node(0, STH) as sth:
            ...             return (await
            ...                 sth.eeprom.read_z_axis_acceleration_offset())
            >>> z_axis_acceleration_offset = run(
            ...     read_z_axis_acceleration_offset())
            >>> isinstance(z_axis_acceleration_offset, float)
            True

        """

        return await self.read_float(address=8, offset=20)

    async def write_z_axis_acceleration_offset(self, offset: int) -> None:
        """Write the acceleration offset of the z-axis to the EEPROM

        Args:

            offset:
                The (negative) offset of the acceleration value in multiples
                of g₀

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from math import isclose
            >>> from icotronic.can.connection import Connection
            >>> from icotronic.can.node.sth import STH

            Write and read the acceleration offset of an STH

            >>> async def write_read_z_axis_acceleration_offset(offset):
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         async with stu.connect_sensor_node(0, STH) as sth:
            ...             await sth.eeprom.write_z_axis_acceleration_offset(
            ...                 offset)
            ...             return (await
            ...                 sth.eeprom.read_z_axis_acceleration_offset())
            >>> acceleration_difference_max = 200
            >>> offset = -(acceleration_difference_max/2)
            >>> offset_read = run(
            ...     write_read_z_axis_acceleration_offset(offset))
            >>> isclose(offset, offset_read)
            True

        """

        await self.write_float(address=8, offset=20, value=offset)


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import run_docstring_examples

    run_docstring_examples(
        STHEEPROM.write_z_axis_acceleration_offset,
        globals(),
        verbose=True,
    )
